from pathlib import Path
from typing import Optional, Tuple


def get_substitutions_datasink_flair(bids_image_id: str) -> list:
    from clinica.pipelines.t1_linear.anat_linear_utils import (  # noqa
        _get_substitutions_datasink,
    )

    return _get_substitutions_datasink(bids_image_id, "FLAIR")


def get_substitutions_datasink_t1_linear(bids_image_id: str) -> list:
    from clinica.pipelines.t1_linear.anat_linear_utils import (  # noqa
        _get_substitutions_datasink,
    )

    return _get_substitutions_datasink(bids_image_id, "T1w")


def _get_substitutions_datasink(bids_image_id: str, suffix: str) -> list:
    """Return file name substitutions for renaming.

    Parameters
    ----------
    bids_image_id : str
        This is the original image BIDS file name without the extension.
        This will be used to get all the BIDS entities that shouldn't
        be modified (subject, session...).

    suffix : str
        The suffix to use for the new file.

    Returns
    -------
    substitutions : List of tuples of str
        List of length 3 containing the substitutions to perform.
    """
    if bids_image_id.endswith(f"_{suffix}"):
        bids_image_id_without_suffix = bids_image_id.removesuffix(f"_{suffix}")
    else:
        raise ValueError(
            f"bids image ID {bids_image_id} should end with provided {suffix}."
        )
    return [
        (
            f"{bids_image_id}Warped_cropped.nii.gz",
            f"{bids_image_id_without_suffix}_space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_{suffix}.nii.gz",
        ),
        (
            f"{bids_image_id}0GenericAffine.mat",
            f"{bids_image_id_without_suffix}_space-MNI152NLin2009cSym_res-1x1x1_affine.mat",
        ),
        (
            f"{bids_image_id}Warped.nii.gz",
            f"{bids_image_id_without_suffix}_space-MNI152NLin2009cSym_res-1x1x1_{suffix}.nii.gz",
        ),
    ]


def print_end_pipeline(anat, final_file):
    """Display end message for <subject_id> when <final_file> is connected."""
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(anat))


def run_n4biasfieldcorrection(
    input_image: Path,
    bspline_fitting_distance: int,
    output_prefix: Optional[str] = None,
    output_dir: Optional[Path] = None,
    save_bias: bool = False,
    verbose: bool = False,
) -> Path:
    """Run n4biasfieldcorrection using antsPy.

    Parameters
    ----------
    input_image : Path
        The path to the input image.

    bspline_fitting_distance : int
        This is the 'spline_param' of n4biasfieldcorrection.

    output_prefix : str, optional
        The prefix to be put at the beginning of the output file names.
        Ex: 'sub-XXX_ses-MYYY'.

    output_dir : Path, optional
        The directory in which to write the output files.
        If not provided, these files will be written in the current directory.

    save_bias : bool, optional
        Whether to save the bias image or not.
        If set to True, the bias image is not returned but saved in the
        provided output_dir with a name of the form '{output_prefix}_bias_image.nii.gz'.
        Default=False.

    verbose : bool, optional
        Control the verbose mode of n4biasfieldcorrection. Set to True can be
        useful for debugging.
        Default=False.

    Returns
    -------
    bias_corrected_output_path : Path
        The path to the bias corrected image.
    """
    from clinica.utils.exceptions import ClinicaMissingDependencyError
    from clinica.utils.stream import cprint, log_and_raise

    try:
        import ants
    except ImportError:
        log_and_raise(
            "The package 'antsPy' is required to run antsRegistration in Python.",
            ClinicaMissingDependencyError,
        )

    output_prefix = output_prefix or ""
    bias_corrected_image = _call_n4_bias_field_correction(
        input_image, bspline_fitting_distance, save_bias=False, verbose=verbose
    )
    if save_bias:
        bias_image = _call_n4_bias_field_correction(
            input_image,
            bspline_fitting_distance,
            save_bias=True,
            verbose=verbose,
        )
        bias_output_path = (
            output_dir or Path.cwd()
        ) / f"{output_prefix}_bias_image.nii.gz"
        ants.image_write(bias_image, str(bias_output_path))
        cprint(f"Writing bias image to {bias_output_path}.", lvl="debug")
    bias_corrected_output_path = (
        output_dir or Path.cwd()
    ) / f"{output_prefix}_bias_corrected_image.nii.gz"
    cprint(
        f"Writing bias corrected image to {bias_corrected_output_path}.", lvl="debug"
    )
    ants.image_write(bias_corrected_image, str(bias_corrected_output_path))

    return bias_corrected_output_path


def _call_n4_bias_field_correction(
    input_image: Path,
    bspline_fitting_distance: int,
    save_bias: bool = False,
    verbose: bool = False,
) -> Path:
    import ants
    from ants.utils.bias_correction import n4_bias_field_correction

    return n4_bias_field_correction(
        ants.image_read(str(input_image)),
        spline_param=bspline_fitting_distance,
        return_bias_field=save_bias,
        verbose=verbose,
    )


def run_ants_registration(
    fixed_image: Path,
    moving_image: Path,
    random_seed: int,
    output_prefix: Optional[str] = None,
    output_dir: Optional[Path] = None,
    verbose: bool = False,
) -> Tuple[Path, Path]:
    """Run antsRegistration using antsPy.

    Parameters
    ----------
    fixed_image : Path
        The path to the fixed image.

    moving_image : Path
        The path to the moving image.

    random_seed : int
        The random seed to be used.

    output_prefix : str, optional
        The prefix to be put at the beginning of the output file names.
        Ex: 'sub-XXX_ses-MYYY'.

    output_dir : Path, optional
        The directory in which to write the output files.
        If not provided, these files will be written in the current directory.

    verbose : bool, optional
        Control the verbose mode of antsRegistration. Set to True can be
        useful for debugging.
        Default=False.

    Returns
    -------
    warped_image_output_path : Path
        The path to the warped nifti image generated by antsRegistration.

    transformation_matrix_output_path : Path
        The path to the transforms to move from moving to fixed image.
        This is a .mat file.

    Raises
    ------
    RuntimeError :
        If results cannot be extracted.
    """
    from clinica.utils.stream import log_and_raise

    registration_results = _call_ants_registration(
        fixed_image, moving_image, random_seed, verbose=verbose
    )
    try:
        warped_image = registration_results["warpedmovout"]
        transformation_matrix = registration_results["fwdtransforms"][-1]
    except (KeyError, IndexError):
        msg = (
            "Something went wrong when calling antsRegistration with the following parameters :\n"
            f"- fixed_image = {fixed_image}\n- moving_image = {moving_image}\n"
            f"- random_seed = {random_seed}\n- type_of_transformation='antsRegistrationSyN[a]'\n"
        )
        log_and_raise(msg, RuntimeError)

    return _write_ants_registration_results(
        warped_image, transformation_matrix, output_prefix or "", output_dir
    )


def _call_ants_registration(
    fixed_image: Path,
    moving_image: Path,
    random_seed: int,
    verbose: bool = False,
) -> dict:
    from clinica.utils.exceptions import ClinicaMissingDependencyError
    from clinica.utils.stream import log_and_raise

    try:
        import ants
    except ImportError:
        log_and_raise(
            "The package 'antsPy' is required to run antsRegistration in Python.",
            ClinicaMissingDependencyError,
        )
    return ants.registration(
        ants.image_read(str(fixed_image)),
        ants.image_read(str(moving_image)),
        type_of_transformation="antsRegistrationSyN[a]",
        random_seed=random_seed,
        verbose=verbose,
    )


def _write_ants_registration_results(
    warped_image,
    transformation_matrix,
    output_prefix: str,
    output_dir: Optional[Path] = None,
) -> Tuple[Path, Path]:
    import shutil

    import ants

    from clinica.utils.stream import cprint

    warped_image_output_path = (
        output_dir or Path.cwd()
    ) / f"{output_prefix}Warped.nii.gz"
    transformation_matrix_output_path = (
        output_dir or Path.cwd()
    ) / f"{output_prefix}0GenericAffine.mat"
    cprint(f"Writing warped image to {warped_image_output_path}.", lvl="debug")
    ants.image_write(warped_image, str(warped_image_output_path))
    shutil.copy(transformation_matrix, transformation_matrix_output_path)

    return warped_image_output_path, transformation_matrix_output_path



from typing import Optional

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.t1_linear.t1_linear as pydra_t1_linear
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-t1-linear"


@clinica_pipeline
@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@option.global_option_group
@option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    n_procs: Optional[int] = None,
) -> None:
    """Affine registration of T1w images to the MNI standard space (Pydra engine).

    Parameters
    ----------
    bids_directory : str
        Path to the input folder containing a BIDS compliant dataset.

    caps_directory : str
        Path to a CAPS folder containing the results of the required pipelines (T1-linear).
        This will also be used for writing the outputs.

    n_procs : int, optional
        Number of processes to use.
    """
    t1_linear_pipeline = pydra_t1_linear.build_core_workflow(
        name="t1-linear-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters={},
    )
    pydra_utils.run(t1_linear_pipeline, n_procs=n_procs)


if __name__ == "__main__":
    cli()



from typing import List

from clinica.pipelines.engine import Pipeline


class T1VolumeParcellation(Pipeline):
    """T1VolumeParcellation - Computation of mean GM concentration for a set of regions.

    Returns:
        A clinica pipeline object containing the T1VolumeParcellation pipeline.
    """

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.atlas import T1AndPetVolumeAtlasName
        from clinica.utils.group import check_group_label

        self.parameters.setdefault("group_label", None)
        check_group_label(self.parameters["group_label"])
        self.parameters.setdefault("atlases", T1AndPetVolumeAtlasName)
        self.parameters.setdefault("modulate", True)

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["file_list", "atlas_list"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return []

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.input_files import t1_volume_template_tpm_in_mni
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import (
            print_groups_in_caps_directory,
            print_images_to_process,
        )

        if not (
            self.caps_directory / "groups" / f"group-{self.parameters['group_label']}"
        ).exists():
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                f"Group {self.parameters['group_label']} does not exist. "
                "Did you run t1-volume or t1-volume-create-dartel pipeline?"
            )

        try:
            gm_mni, _ = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                t1_volume_template_tpm_in_mni(
                    group_label=self.parameters["group_label"],
                    tissue_number=1,
                    modulation=self.parameters["modulate"],
                ),
            )
        except ClinicaException as e:
            final_error_str = "Clinica faced error(s) while trying to read files in your CAPS directory.\n"
            final_error_str += str(e)
            raise ClinicaCAPSError(final_error_str)

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )
        read_parameters_node.inputs.file_list = gm_mni
        read_parameters_node.inputs.atlas_list = self.parameters["atlases"]

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint("The pipeline will last a few seconds per image.")

        self.connect(
            [
                (read_parameters_node, self.input_node, [("file_list", "file_list")]),
                (read_parameters_node, self.input_node, [("atlas_list", "atlas_list")]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from ..t1_volume_parcellation import (
            t1_volume_parcellation_utils as parcellation_utils,
        )

        atlas_stats_node = npe.MapNode(
            nutil.Function(
                input_names=["in_image", "atlas_list"],
                output_names=["atlas_statistics"],
                function=parcellation_utils.atlas_statistics,
            ),
            name="atlas_stats_node",
            iterfield=["in_image"],
        )
        outputnode = npe.Node(
            nutil.IdentityInterface(fields=["atlas_statistics"]),
            name="outputnode",
            mandatory_inputs=True,
        )

        datasink = npe.Node(nio.DataSink(), name="datasink")

        datasink.inputs.base_directory = str(self.caps_directory)
        datasink.inputs.parameterization = True
        datasink.inputs.regexp_substitutions = [
            (
                r"(.*)(atlas_statistics)/.*/(sub-(.*)_ses-(.*)_T1.*)$",
                r"\1/subjects/sub-\4/ses-\5/t1/spm/dartel/group-"
                + self.parameters["group_label"]
                + r"/\2/\3",
            )
        ]
        self.connect(
            [
                (self.input_node, atlas_stats_node, [("file_list", "in_image")]),
                (self.input_node, atlas_stats_node, [("atlas_list", "atlas_list")]),
                (
                    atlas_stats_node,
                    outputnode,
                    [("atlas_statistics", "atlas_statistics")],
                ),
                (outputnode, datasink, [("atlas_statistics", "atlas_statistics")]),
            ]
        )



from os import PathLike
from pathlib import Path
from typing import List, Tuple, Union

import nibabel as nib
import numpy as np
from nibabel import Nifti1Header

__all__ = [
    "apply_binary_mask",
    "compute_atlas_statistics",
    "create_binary_mask",
    "create_pvc_mask",
    "get_from_list",
    "init_input_node",
    "normalize_to_reference",
    "build_pet_pvc_name",
]


def init_input_node(pet_nii: str) -> str:
    from clinica.utils.filemanip import get_subject_id, load_volume
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    image_id = get_subject_id(pet_nii)
    try:
        load_volume(pet_nii)
    except ValueError as e:
        error_msg = f"Clinica could not load volumes for {image_id.replace('_', ' | ')}. {str(e)}"
        cprint(error_msg, lvl="error")
        raise ValueError(error_msg)
    print_begin_image(image_id)

    return pet_nii


def _check_non_empty_tissue_list(tissues: List[PathLike]) -> None:
    """Check that provided list is non-empty."""
    if len(tissues) == 0:
        raise RuntimeError(
            "The length of the list of tissues must be greater than zero."
        )


def _aggregate_tissue_images(
    tissues: List[PathLike],
) -> Tuple[np.ndarray, np.ndarray, Nifti1Header]:
    """Aggregates the image data contained in the tissue images provided.

    Parameters
    ----------
    tissues : list of PathLike
        List of tissue images to aggregate.

    Returns
    -------
    data : np.ndarray
        Aggregated data.

    affine : np.ndarray
        Affine of the first image, acting as the affine
        of the aggregated image under the assumption that
        all images have the same affine.

    header : Nifti1Header
        Header of the aggregated image.
    """
    _check_non_empty_tissue_list(tissues)
    first_image = nib.load(tissues[0])
    data = np.zeros(shape=list(first_image.get_fdata(dtype="float32").shape))
    for image in tissues:
        data += nib.load(image).get_fdata(dtype="float32")
    return data, first_image.affine, first_image.header


def create_binary_mask(tissues: List[PathLike], threshold: float = 0.3) -> Path:
    """Create a binary mask Nifti1Image from the list of tissues.

    Tissue images are summed and the result is thresholded with the
    provided `threshold` input.

    Parameters
    ----------
    tissues : list of PathLike
        List of paths to tissue Nifti1Images. Must be non-empty.

    threshold : float, optional
        Threshold to apply when binarizing the Nifti1Image.
        Default=0.3.

    Returns
    -------
    out_mask : Path
        The path to the binary mask Nifti1Image as a string.
    """
    from clinica.utils.filemanip import get_filename_no_ext

    data, affine, header = _aggregate_tissue_images(tissues)
    data = (data > threshold) * 1.0
    out_mask = Path.cwd() / f"{get_filename_no_ext(tissues[0])}_brainmask.nii"
    mask = nib.Nifti1Image(data, affine, header=header)
    nib.save(mask, out_mask)

    return out_mask


def apply_binary_mask(image: Path, binary_mask: Path) -> Path:
    """Apply the provided `binary_mask` to the provided `image`.

    Parameters
    ----------
    image : Path
        The path to the Nifti1Image to apply the mask on.

    binary_mask : Path
        The path to the Nifti1Image containing the mask.

    Returns
    -------
    masked_image_path : Path
        The path to the masked Nifti1Image.
    """
    import nibabel as nib

    original_image = nib.load(image)
    mask = nib.load(binary_mask)
    data = original_image.get_fdata(dtype="float32") * mask.get_fdata(dtype="float32")
    masked_image_path = Path.cwd() / f"masked_{image.name}"
    masked_image = nib.Nifti1Image(
        data, original_image.affine, header=original_image.header
    )
    nib.save(masked_image, masked_image_path)

    return masked_image_path


def create_pvc_mask(tissues: List[PathLike]) -> Path:
    """Create a pvc mask from tissue list.

    Parameters
    ----------
    tissues : list of PathLike
        List of paths to tissue Nifti1Images. Must be non-empty.

    Returns
    -------
    out_mask : Path
        The path to the resulting mask Nifti1Image.
    """
    background, affine, header = _aggregate_tissue_images(tissues)
    shape = background.shape
    shape += tuple([len(tissues) + 1])
    data = np.empty(shape=shape, dtype=np.float64)
    for i, tissue in enumerate(tissues):
        image = nib.load(tissue)
        data[..., i] = np.array(image.get_fdata(dtype="float32"))
    background = 1.0 - background
    data[..., len(tissues)] = np.array(background)
    out_mask = Path.cwd() / "pvc_mask.nii"
    mask = nib.Nifti1Image(data, affine, header=header)
    nib.save(mask, out_mask)

    return out_mask


def build_pet_pvc_name(pet_image: Union[str, PathLike], pvc_method: str) -> str:
    """Build the name for the PET PVC interface.

    Parameters
    ----------
    pet_image : str or Path
        The path to the PET scan file.

    pvc_method : str
        Name of the PVC method. This will be concatenated
        with the `pet_image` filename.

    Returns
    -------
    pet_pvc_path : str
        Name for the PET PVC interface.

    Examples
    --------
    >>> build_pet_pvc_name(
    ...     "/home/bids/sub-01/ses-M00/pet/sub-01_ses-M00_task-rest_trc-av45_pet.nii.gz",
    ...     "RBV"
    ...)
    'pvc-rbv_sub-01_ses-M00_task-rest_trc-av45_pet.nii.gz'

    """
    pet_image = Path(pet_image)

    return f"pvc-{pvc_method.lower()}_{pet_image.name}"


def normalize_to_reference(pet_image: Path, region_mask: Path) -> Path:
    """Normalize the provided `pet_image` by dividing by the mean
    value of the region defined by the provided `region_mask`.

    Parameters
    ----------
    pet_image : Path
        The path to the Nifti1Image which should be normalized.

    region_mask : Path
        The path to the mask to be used to define the region.

    Returns
    -------
    suvr_pet_path : Path
        The path to the normalized Nifti1Image.
    """
    pet = nib.load(pet_image)
    ref = nib.load(region_mask)
    region = pet.get_fdata(dtype="float32") * ref.get_fdata(dtype="float32")
    region_mean = np.nanmean(np.where(region != 0, region, np.nan))
    data = pet.get_fdata(dtype="float32") / region_mean
    suvr_pet_path = Path.cwd() / f"suvr_{pet_image.name}"
    suvr_pet = nib.Nifti1Image(data, pet.affine, header=pet.header)
    nib.save(suvr_pet, suvr_pet_path)

    return suvr_pet_path


def compute_atlas_statistics(image: Path, atlas_names: List[str]) -> List[Path]:
    """Generate regional measure from atlas_list in TSV files.

    For each atlas name provided it calculates for the input image the mean
    for each region in the atlas and saves it to a TSV file.

    Parameters
    ----------
    image : Path
        The path to the Nifti image.

    atlas_names : List of str
        List of names of atlas to be applied.

    Returns
    -------
    atlas_statistics : List of paths
        List of paths to TSV files.
    """
    from clinica.utils.filemanip import get_filename_no_ext
    from clinica.utils.statistics import statistics_on_atlas

    atlas_statistics_list = []
    for atlas in atlas_names:
        out_atlas_statistics = (
            Path.cwd() / f"{get_filename_no_ext(image)}_space-{atlas}_statistics.tsv"
        )
        statistics_on_atlas(image, atlas, out_atlas_statistics)
        atlas_statistics_list.append(out_atlas_statistics)
        break  # Why is there a break here ????
    return atlas_statistics_list


def get_from_list(in_list: list, index: int):
    return in_list[index]



from pathlib import Path
from typing import Callable

from ._statistics import StatisticsResults

__all__ = ["StatisticsResultsSerializer"]


class StatisticsResultsSerializer:
    """This class is responsible for writing instances of `StatisticsResults`
    to disk through different methods.

    Attributes
    ----------
    output_file : PathLike
        Path and filename root to be used.
    """

    def __init__(self, output_file: Path):
        self.output_file = output_file
        self.json_extension = "_results.json"
        self.json_indent = 4
        self.mat_extension = ".mat"

    def save(self, result: StatisticsResults, method: str) -> None:
        """Save provided `StatisticsResults` to disk with provided method.

        Parameters
        ----------
        result : StatisticsResults
            Results to be saved.

        method : str
            Name of the saving method to use.
        """
        writer = self._get_writer(method)
        writer(result)

    def _get_writer(self, method: str) -> Callable[[StatisticsResults], None]:
        """Returns a writer method from its name.

        Parameters
        ----------
        method : str
            The name of the writing method to use.

        Returns
        -------
        Callable :
            The writing method.
        """
        if method.lower() == "json":
            return self._write_to_json
        elif method.lower() == "mat":
            return self._write_to_mat
        raise NotImplementedError(f"Serializing method {method} is not implemented.")

    def _write_to_json(self, results: StatisticsResults) -> None:
        """Write the provided `StatisticsResults` to JSON format.

        Parameters
        ----------
        results : StatisticsResults
            The results to write to disk in JSON format.
        """
        import json
        import os

        from clinica.utils.stream import cprint

        out_json_file = Path(str(self.output_file) + self.json_extension)
        if not os.path.exists(out_json_file.parents[0]):
            os.makedirs(out_json_file.parents[0])
        cprint(
            msg=f"Writing results to JSON in {out_json_file}...",
            lvl="info",
        )
        with open(out_json_file, "w") as fp:
            json.dump(results.to_json(indent=self.json_indent), fp)

    def _write_to_mat(self, results: StatisticsResults) -> None:
        """Write the provided `StatisticsResults` to MAT format.

        Parameters
        ----------
        results : StatisticsResults
            The results to write to disk in MAT format.
        """
        from scipy.io import savemat

        from clinica.utils.stream import cprint

        # These labels are used for compatibility with the previous
        # MATLAB implementation of the Statistics Surface Pipeline
        # of Clinica.
        struct_labels = {
            "coefficients": "coef",
            "TStatistics": "tvaluewithmask",
            "uncorrectedPValue": "uncorrectedpvaluesstruct",
            "correctedPValue": "correctedpvaluesstruct",
            "FDR": "FDR",
        }
        for name, res in results.to_dict().items():
            if name in struct_labels:
                mat_filename = str(self.output_file) + "_" + name + self.mat_extension
                cprint(
                    msg=f"Writing {name} results to MAT in  {mat_filename}",
                    lvl="info",
                )
                savemat(mat_filename, {struct_labels[name]: res})



from __future__ import annotations

from pathlib import PurePath

from pydra.mark import annotate, task


def download_file(url: str, to: str) -> PurePath:
    from shutil import copyfileobj
    from ssl import SSLContext
    from urllib.request import urlopen

    print(f"Downloading {url} to {to}...")

    response = urlopen(url=url, context=SSLContext())
    with open(to, mode="wb") as f:
        copyfileobj(response, f)

    return PurePath(to)


@task
@annotate({"return": {"mni_template_file": PurePath}})
def download_mni_template_2009a() -> PurePath:
    return _download_mni_template(version="09a")


@task
@annotate({"return": {"mni_template_file": PurePath}})
def download_mni_template_2009c() -> PurePath:
    return _download_mni_template(version="09c")


def _download_mni_template(version: str) -> PurePath:
    from pathlib import Path

    return download_file(
        url=f"https://aramislab.paris.inria.fr/files/data/img_t1_linear/mni_icbm152_t1_tal_nlin_sym_{version}.nii.gz",
        to=str(Path.cwd() / f"mni_icbm152_t1_tal_nlin_sym_{version}.nii.gz"),
    )


@task
@annotate({"return": {"ref_template_file": PurePath}})
def download_ref_template() -> PurePath:
    from pathlib import Path

    return download_file(
        url="https://aramislab.paris.inria.fr/files/data/img_t1_linear/ref_cropped_template.nii.gz",
        to=str(Path.cwd() / "ref_cropped_template.nii.gz"),
    )



"""Convert the UKB dataset into BIDS."""

from typing import Optional

from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: UserProvidedPath,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
):
    """Convert the entire dataset to BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """
    from clinica.iotools.bids_utils import StudyName, write_modality_agnostic_files
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.check_dependency import ThirdPartySoftware, check_software
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path
    from .ukb_utils import (
        find_clinical_data,
        merge_imaging_and_clinical_data,
        prepare_dataset_to_bids_format,
        read_imaging_data,
        write_bids,
    )

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    check_software(ThirdPartySoftware.DCM2NIIX)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.UKB)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    if n_procs != 1:
        cprint(
            f"{get_converter_name(StudyName.UKB)} converter does not support multiprocessing yet. n_procs set to 1.",
            lvl="warning",
        )
    result = prepare_dataset_to_bids_format(
        merge_imaging_and_clinical_data(
            read_imaging_data(path_to_dataset),
            find_clinical_data(path_to_clinical),
        )
    )
    write_bids(
        to=bids_dir,
        participants=result["participants"],
        sessions=result["sessions"],
        scans=result["scans"],
        dataset_directory=path_to_dataset,
    )
    readme_data = {
        "link": "https://www.ukbiobank.ac.uk/",
        "desc": (
            "UK Biobank is a large-scale biomedical database and research resource, containing in-depth genetic and "
            "health information from half a million UK participants. The database is regularly augmented with "
            "additional data and is globally accessible to approved researchers undertaking vital research into the "
            "most common and life-threatening diseases. It is a major contributor to the advancement of modern "
            "medicine it and has led to the discovery of several scientific advances and numerous treatments to "
            "improve human health."
        ),
    }
    write_modality_agnostic_files(
        study_name=StudyName.UKB,
        readme_data=readme_data,
        bids_dir=bids_dir,
    )
    cprint("Conversion to BIDS succeeded.", lvl="info")



"""This module contains Nipype tasks used in multiple DWI preprocessing pipelines.

Nipype tasks must be 'self-contained' such that only primitive type hints can be
used. The tasks are simple wrappers around a properly implemented Python function.
"""


def generate_index_file_task(
    b_values_filename: str,
    image_id=None,
    output_dir=None,
) -> str:
    """Wrapper around 'generate_index_file' for Nipype."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.utils import generate_index_file

    return str(
        generate_index_file(
            Path(b_values_filename),
            image_id,
            Path(output_dir) if output_dir else None,
        )
    )


def generate_acq_file_task(
    dwi_filename: str,
    fsl_phase_encoding_direction: str,
    total_readout_time: str,
    image_id=None,
    output_dir=None,
) -> str:
    """Wrapper around 'generate_acq_file' for Nipype."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.utils import generate_acq_file

    return str(
        generate_acq_file(
            Path(dwi_filename),
            fsl_phase_encoding_direction,
            total_readout_time,
            image_id,
            Path(output_dir) if output_dir else None,
        )
    )


def compute_average_b0_task(
    dwi_filename: str,
    b_value_filename: str,
    b_value_threshold: float = 5.0,
    squeeze: bool = False,
    out_file: str = None,
) -> str:
    """Nipype task for 'compute_average_b0'."""
    from clinica.pipelines.dwi.preprocessing.utils import check_file, compute_average_b0

    return str(
        compute_average_b0(
            check_file(dwi_filename),
            check_file(b_value_filename),
            b_value_threshold,
            squeeze,
            out_file,
        )
    )



from pathlib import Path

__all__ = [
    "init_input_node",
    "save_to_caps",
    "write_tsv_files",
]


def init_input_node(
    t1w: Path, recon_all_args: str, output_dir: Path
) -> tuple[str, Path, str, Path]:
    """Initialize the pipeline.

    This function will:
        - Extract <image_id> (e.g. sub-CLNC01_ses-M000) T1w filename;
        - Check FOV of T1w;
        - Create SUBJECTS_DIR for recon-all (otherwise, the command won't run);
        - Print begin execution message.
    """
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    image_id = get_subject_id(t1w)
    flags = _check_flags(t1w, recon_all_args)
    subjects_dir = output_dir / image_id
    subjects_dir.mkdir(parents=True, exist_ok=True)
    print_begin_image(image_id, ["ReconAllArgs"], [flags])

    return image_id, t1w, flags, subjects_dir


def _check_flags(t1w_image: Path, recon_all_args: str) -> str:
    """Check `recon_all_args` flags for `in_t1w` image.

    Currently, this function only adds '-cw256' if the FOV of `in_t1w` is greater than 256.

    Parameters
    ----------
    t1w_image : str or Path
        The path to the T1W image.

    recon_all_args : str
        ??????.

    Returns
    -------
    str :
        ????.
    """
    import nibabel as nib

    image = nib.load(t1w_image)
    if any(
        [
            v * t > 256
            for v, t in zip(image.header.get_zooms(), image.header.get_data_shape())
        ]
    ):
        return " ".join([recon_all_args, "-cw256"])
    return recon_all_args


def write_tsv_files(subjects_dir: Path, image_id: str) -> str:
    """
    Generate statistics TSV files in `subjects_dir`/regional_measures folder for `image_id`.

    Notes
    -----
    We do not need to check the line "finished without error" in scripts/recon-all.log.
    If an error occurs, it will be detected by Nipype and the next nodes (including
    write_tsv_files will not be called).
    """
    from clinica.pipelines.anatomical.freesurfer.utils import generate_regional_measures
    from clinica.utils.stream import cprint

    if (subjects_dir / image_id / "mri" / "aparc+aseg.mgz").is_file():
        generate_regional_measures(subjects_dir, image_id)
    else:
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')} does not contain "
                f"mri/aseg+aparc.mgz file. Creation of regional_measures/ folder will be skipped."
            ),
            lvl="warning",
        )
    return image_id


def save_to_caps(
    source_dir: Path, image_id: str, caps_dir: Path, overwrite_caps: bool = False
) -> str:
    """Save `source_dir`/`image_id`/ to CAPS folder.

    This function copies FreeSurfer segmentation and regional_measures folder of `source_dir`/`image_id`/ to
    `caps_dir`/subjects/<participant_id>/<session_id>/t1_freesurfer_cross_sectional/
    where `image_id` = <participant_id>_<session_id>.

    The `source_dir`/`image_id`/ folder should contain the following elements:
        - fsaverage, lh.EC_average and rh.EC_average symbolic links automatically generated by recon-all
        - `image_id`/ folder containing the FreeSurfer segmentation
        - regional_measures/ folder containing TSV files

    Notes
    -----
    We do not need to check the line "finished without error" in scripts/recon-all.log.
    If an error occurs, it will be detected by Nipype and the next nodes (i.e.  save_to_caps will not be called).
    """
    import shutil

    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_end_image

    participant_id = image_id.split("_")[0]
    session_id = image_id.split("_")[1]
    destination_dir = (
        caps_dir.expanduser()
        / "subjects"
        / participant_id
        / session_id
        / "t1"
        / "freesurfer_cross_sectional"
    )
    source_file = (
        source_dir.expanduser() / image_id / image_id / "mri" / "aparc+aseg.mgz"
    )
    destination_file = destination_dir / image_id / "mri" / "aparc+aseg.mgz"
    if source_file.is_file():
        if destination_file.is_file():
            if overwrite_caps:
                shutil.rmtree(destination_dir)
                shutil.copytree(
                    src=source_dir / image_id / image_id,
                    dst=destination_dir / image_id,
                    symlinks=True,
                )
                shutil.copytree(
                    src=source_dir / image_id / "regional_measures",
                    dst=destination_dir / "regional_measures",
                    symlinks=True,
                )
        else:
            shutil.copytree(
                src=source_dir / image_id / image_id,
                dst=destination_dir / image_id,
                symlinks=True,
            )
            shutil.copytree(
                src=source_dir / image_id / "regional_measures",
                dst=destination_dir / "regional_measures",
                symlinks=True,
            )
        print_end_image(image_id)
    else:
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')} does not "
                "contain mri/aseg+aparc.mgz file. Copy will be skipped."
            ),
            lvl="warning",
        )
    return image_id



from os import PathLike
from typing import Optional

import click

from clinica.iotools.converters import cli_param


@click.command(name="ixi-to-bids")
@cli_param.dataset_directory
@cli_param.bids_directory
@cli_param.clinical_data_directory
@cli_param.subjects_list
def cli(
    dataset_directory: PathLike,
    bids_directory: PathLike,
    clinical_data_directory: PathLike,
    subjects_list: Optional[PathLike] = None,
) -> None:
    """IXI to BIDS converter."""
    from .ixi_to_bids import convert

    convert(dataset_directory, bids_directory, clinical_data_directory, subjects_list)


if __name__ == "__main__":
    cli()



import warnings
from dataclasses import dataclass
from enum import Enum
from functools import partial
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import pandas as pd

from clinica.utils.pet import Tracer

__all__ = ["Modality", "paths_to_bids"]


class Modality(str, Enum):
    T1 = "t1"
    AV45 = "av45"
    FLUTE = "flute"
    PIB = "pib"

    @property
    def name_of_path(self) -> str:
        return "Path_to_T1" if self == "t1" else "Path_to_pet"

    @property
    def tracer(self) -> Tracer:
        if self == "t1":
            raise ValueError("T1 modality does not have a tracer.")
        if self == "flute":
            return Tracer.FMM
        if self == "pib":
            return Tracer.PIB
        if self == "av45":
            return Tracer.AV45

    @property
    def bids_folder(self) -> str:
        return "anat" if self == "t1" else "pet"

    @property
    def suffix(self) -> str:
        return "T1w" if self == "t1" else f"trc-{self.tracer.value}_pet"


@dataclass
class Scan:
    """Class for keeping track of the relationship between a subject ID,
    a session ID, and a path to a modality scan.
    """

    subject_id: str
    session_id: str
    path: Path


@dataclass
class ScanCollection:
    """
    Maintain coherence in the collection : it is impossible to add
    a scan for an already existing subject session pair.
    """

    scans: List[Scan]
    modality: Modality

    def _is_new_scan_valid(self, new_scan: Scan) -> bool:
        for scan in self.scans:
            if (scan.subject_id, scan.session_id) == (
                new_scan.subject_id,
                new_scan.session_id,
            ):
                warnings.warn(
                    f"Already a scan for {self.modality} in the collection for subject {scan.subject_id} "
                    f"and session {scan.session_id} : {scan.path}. "
                    f"Cannot add another scan {new_scan.path} for this subject session pair."
                )
                return False
        return True

    def add(self, scan: Scan):
        if self._is_new_scan_valid(scan):
            self.scans.append(scan)

    def add_scan(self, subject_id: str, session_id: str, path: Path):
        """Add a scan to the collection."""
        self.add(Scan(subject_id, session_id, path))

    def to_df(self) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "Subjects_ID": [scan.subject_id for scan in self.scans],
                "Session_ID": [scan.session_id for scan in self.scans],
                self.modality.name_of_path: [str(scan.path) for scan in self.scans],
            }
        )


def paths_to_bids(
    path_to_dataset: Path,
    path_to_csv: Path,
    bids_dir: Path,
    modality: Modality,
    overwrite: bool = False,
    n_procs: Optional[int] = 1,
) -> List[Optional[Path]]:
    """Convert all the T1 images found in the AIBL dataset downloaded in BIDS.

    Parameters
    ----------
    path_to_dataset : Path
        The path to the AIBL dataset.

    path_to_csv : Path
        The path to the csv file containing clinical data.

    bids_dir : Path
        The path to save the AIBL-T1-dataset converted in a BIDS format.

    modality : Modality
        The modality to convert.

    overwrite : bool
        If True previous existing outputs will be erased.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.

    Return
    ------
    list of Path or None :
        A list of paths to all the images that are potentially converted in a BIDS format
        and saved in the bids_dir.
        Elements of the list can be None if the image doesn't exist.
    """
    from functools import partial
    from multiprocessing import Pool

    (bids_dir / "conversion_info").mkdir(parents=True, exist_ok=True)

    if modality == Modality.T1:
        images = _find_path_to_t1(path_to_dataset, path_to_csv)
    else:
        images = _find_path_to_pet_modality(path_to_dataset, path_to_csv, modality)

    images.to_csv(
        bids_dir / "conversion_info" / f"{modality.value}_paths.tsv",
        index=False,
        sep="\t",
        encoding="utf-8",
    )
    images_list = list([data for _, data in images.iterrows()])
    create_file_ = partial(
        _create_file,
        modality=modality,
        bids_dir=bids_dir,
        overwrite=overwrite,
    )
    # If n_procs==1 do not rely on a Process Pool to enable classical debugging
    if n_procs == 1:
        return [create_file_(image) for image in images_list]
    with Pool(processes=n_procs) as pool:
        output_file_treated = pool.map(create_file_, images_list)

    return output_file_treated


def _find_path_to_t1(path_to_dataset: Path, path_to_csv: Path) -> pd.DataFrame:
    """Creates a DataFrame for the T1 images.
    For each image, the subject ID, the session ID and the path to the image are
    reported.

    Parameters
    ----------
    path_to_dataset : Path
        Path to AIBL dataset.

    path_to_csv : Path
        Path to the folder with the CSV files downloaded.

    Return
    ------
    pd.DataFrame :
        Pandas dataframe which contains all the paths for the T1
        images, and the corresponding subject ID, and session ID.
    """
    import pandas as pd

    # two csv_files contain information regarding the T1w MRI images
    mri_dfs = tuple(
        [
            pd.read_csv(_get_first_file_matching_pattern(path_to_csv, pattern))
            for pattern in ("aibl_mrimeta_*.csv", "aibl_mri3meta_*.csv")
        ]
    )
    subjects_id = _listdir_nohidden(path_to_dataset)
    # all the subjects downloaded are taken into account for the conversion, except this sample
    if "0151083" in subjects_id:
        del subjects_id[subjects_id.index("0151083")]
    t1_scans = _find_path_to_t1_adni(mri_dfs, subjects_id, path_to_dataset)
    t1_scans = _find_path_to_t1_sag(subjects_id, path_to_dataset, t1_scans)
    return t1_scans.to_df()


def _get_first_file_matching_pattern(folder: Path, pattern: str) -> Path:
    """Return the first file in given folder matching the provided pattern.
    Raise a ValueError if no such file found.
    """
    if pattern == "":
        raise ValueError("Pattern is not valid.")
    try:
        return sorted([f for f in folder.glob(pattern)])[0]
    except IndexError:
        raise ValueError(f"No file matching pattern {folder}/{pattern}.")


def _listdir_nohidden(folder: Path) -> List[str]:
    """List the names of all the subdirectories of provided folder path, except the hidden folders.

    Parameters
    ----------
    folder: Path
        The path to list subdirectories from.

    Returns
    -------
    list of str:
        Names of subdirectories found within provided folder.
    """
    return [
        str(file.name)
        for file in folder.iterdir()
        if file.is_dir() and not file.name.startswith(".")
    ]


def _find_path_to_t1_adni(
    mri_files: Tuple[pd.DataFrame, pd.DataFrame],
    subject_ids: List[str],
    path_to_dataset: Path,
) -> ScanCollection:
    """Creates a Dataframe which contains all the paths to the T1
    images which are ADNI compliant (as explained in the AIBL website).
    These images differ from the others T1 of the dataset since the exam
    date is reported in the CSV file.

    Parameters
    ----------
    mri_files : tuple of 2 DataFrames
        List of dataframes extracted from the clinical data.
        Indeed, there are two files which describe the parameters of
        the T1 images : MRI 1.5 T and MRI 3T.

    subject_ids : list of str
        List of subjects IDs in the downloaded dataset.

    path_to_dataset : Path
        Path to AIBL dataset.

    Return
    ------
    ScanCollection :
        The collected T1 scan paths.
    """
    scans = ScanCollection([], Modality.T1)

    for subject_id in subject_ids:
        for mri_file in mri_files:
            if int(subject_id) in list(mri_file.RID):
                path_to_subject = path_to_dataset / str(subject_id)
                for folder in _listdir_nohidden(path_to_subject):
                    if path_to_t1_folder := _find_t1_folder(folder, path_to_subject):
                        for exam_date in _listdir_nohidden(path_to_t1_folder):
                            session_id = _find_patient_session_id(
                                exam_date, subject_id, mri_file
                            )
                            if session_id != "-4":
                                path_to_t1 = path_to_t1_folder / str(exam_date)
                                for image_id in _listdir_nohidden(path_to_t1):
                                    scans.add_scan(
                                        subject_id, session_id, path_to_t1 / image_id
                                    )
    return scans


def _find_t1_in_paths(
    subdirectory: str,
    path_to_t1_images: Path,
    paths_to_convert: Sequence[str],
) -> Optional[Path]:
    """Find the directory containing T1 images.

    Parameters
    ----------
    subdirectory : str
        Name of the folder.

    path_to_t1_images : Path
        Path to T1 images.

    paths_to_convert : Sequence[str]
        Sequence of paths to convert.

    Return
    ------
    str or None:
        Previous path to arrive to the T1 image.
        Return None if no path found.
    """
    for folder in paths_to_convert:
        if folder == subdirectory:
            return path_to_t1_images / subdirectory
    return None


_find_t1_folder = partial(
    _find_t1_in_paths,
    paths_to_convert=[
        "MPRAGE_ADNI_confirmed",
        "MPRAGE",
        "MPRAGE_ADNI_confirmed_RPT",
        "MPRAGE_ADNI_confirmed_REPEATX2",
        "MPRAGE_ADNI_confirmed_repeat",
        "MPRAGE_ADNI_confirmed_REPEAT",
        "MPRAGE_ADNI_conf_REPEAT",
    ],
)


def _find_patient_session_id(
    exam_date: str, subject_id: str, metadata_df: pd.DataFrame
) -> str:
    """Returns the session_ID for the requested subject.

    It controls if the dates corresponding to the image (from the name of the subdirectory)
    correspond to one of the dates listed from the csv_file for the subject analysed.
    The session_ID is the corresponding session for that patient in that date.

    It returns -4 if there are no information.

    Parameters
    ----------
    exam_date : str
        Date where the image has been taken, it is saved from the name
        of the corresponding subdirectory.

    subject_id : str
        The subject's identifier.

    metadata_df : pd.DataFrame
        DataFrame obtained from the CSV file where all the information are listed.

    Return
    ------
    str :
        Session ID of the patient.
    """
    import re

    session_id = None
    indices = _find_correspondence_index(subject_id, metadata_df)
    csv_date = metadata_df.EXAMDATE[indices]

    for index in indices:
        if str(csv_date[index]) == "-4":
            continue
        m = re.search(
            "([0-9].*)-(.*)-(.*)_(.*)_(.*)_(.*)", exam_date
        )  # string from image directory
        p = re.search(
            "(.*)/(.*)/(.*)", str(csv_date[index])
        )  # string from the date of the csv_file
        if (
            (p.group(1) == m.group(2))
            & (p.group(2) == m.group(3))
            & (p.group(3) == m.group(1))
        ):
            session_id = metadata_df.VISCODE[index]

    return session_id or "-4"


def _find_correspondence_index(subject_id: str, df: pd.DataFrame) -> List[str]:
    """Returns the index of the CSV file analysed for a given subject.

    Parameters
    ----------
    subject_id : str
        The subject's identifier.

    df : pd.DataFrame
        The dataframe in which to look for subject data.

    Return
    ------
    list of str :
        List of indexes for the subject

    Raises
    ------
    ValueError :
        If no correspondence was found.
    """
    for rid in df.RID:
        if subject_id == str(rid):
            return df.RID[df.RID == rid].index.tolist()
    raise ValueError(
        f"Unable to find correspondence between the subject ID {subject_id} "
        "and RID in provided dataframe."
    )


def _find_path_to_t1_sag(
    subjects_id: List[str],
    path_to_dataset: Path,
    t1_scans: ScanCollection,
) -> ScanCollection:
    """Creates a ScanCollection which contains all the paths to the T1 images which are not
    ADNI compliant, they contain the word "SAG" in their name.

    Parameters
    ----------
    subjects_id : list of str
        Subjects ID in the downloaded dataset.

    path_to_dataset : Path
        Path to AIBL dataset.

    t1_scans : ScanCollection
        The previously computed collection of T1 scan paths.
        The function will potentially add new scan paths to this collection.

    Return
    ------
    ScanCollection :
        The collected T1 scan paths.

    Notes
    -----
    This function completes the list of all the T1 paths including all the images where
    we didn't find the exam data, but we can fix it with further analysis.
    """
    for subject_id in subjects_id:
        path_to_t1 = path_to_dataset / str(subject_id)
        directories_to_analyze = (
            "MPRAGESAGISOp2ND",
            "MPRAGE_SAG_ISO_p2_ND",
            "MPRAGE_SAG_ISO_p2",
        )
        subdirectory_for_subject = [
            d for d in _listdir_nohidden(path_to_t1) if d in directories_to_analyze
        ]
        if not subdirectory_for_subject:
            continue
        path_to_t1 = path_to_t1 / subdirectory_for_subject[0]
        exam_date = _listdir_nohidden(path_to_t1)
        session_id = "M54" if subject_id in (342, 557) else "M00"
        # if for a subject in the same session we have both this image
        # and the "ADNI" compliant we are converting the second one
        # since the exam date is more precise.
        path_to_t1 = path_to_t1 / str(exam_date[0])
        image_id = _listdir_nohidden(path_to_t1)
        path_to_t1 = path_to_t1 / image_id[0]
        t1_scans.add_scan(subject_id, session_id, path_to_t1)

    return t1_scans


def _find_path_to_pet_modality(
    path_to_dataset: Path,
    path_to_csv: Path,
    modality: Modality,
) -> pd.DataFrame:
    """Create a Dataframe which contains all the paths to the PET
    image of a modality (for example AV45 or PIB).

    Parameters
    ----------
    path_to_dataset : Path
        Path to AIBL dataset.

    path_to_csv : Path
        Path to CSV file.

    modality : Modality
        The PET modality for which to find paths.

    Return
    ------
    pd.DataFrame :
        A dataframe which contains the path for PET images for a single
        modality and subject_ID and session_ID are reported for each path.
    """
    df = _load_pet_csv(path_to_csv, modality)
    subjects_id = _listdir_nohidden(path_to_dataset)
    scans = ScanCollection([], modality)

    if "0151083" in subjects_id:
        del subjects_id[subjects_id.index("0151083")]

    for subject_id in subjects_id:
        if int(subject_id) in list(df.RID):
            path_to_subject = path_to_dataset / str(subject_id)
            subject_subdirectories_pet = _check_subdirectories_pet(
                _listdir_nohidden(path_to_subject),
                _list_folder_without_pet(),
            )
            for folder in subject_subdirectories_pet:
                path_to_pet_folder = path_to_subject / folder
                exam_dates = _listdir_nohidden(path_to_pet_folder)
                # exam date of the image which is going to be converted
                for exam_date in exam_dates:
                    # selection of the session_ID matching the data in the csv_file with the one of the image
                    session_id = _find_patient_session_id(
                        str(exam_date), subject_id, df
                    )
                    if session_id != "-4":
                        path_to_pet = path_to_pet_folder / str(exam_date)
                        # For the RID 1607 there are two PET images of the flute modality, and we select the first
                        if (
                            subject_id == "1607"
                            and folder == "Flute_256_1.6_Zoom_plain_4_x_4_Iter"
                        ):
                            image_ids = ["I442930"]
                        else:
                            image_ids = _listdir_nohidden(path_to_pet)
                        for image_id in image_ids:
                            scans.add_scan(
                                subject_id, session_id, path_to_pet / image_id
                            )
    return scans.to_df()


def _load_pet_csv(path_to_csv: Path, modality: Modality) -> pd.DataFrame:
    import pandas as pd

    path_to_csv_pet_modality = _get_first_file_matching_pattern(
        path_to_csv, f"aibl_{modality.value}meta_*.csv"
    )

    if not path_to_csv_pet_modality.exists():
        raise FileNotFoundError(
            f"{path_to_csv_pet_modality} file not found in clinical data folder."
        )
    # Latest version of Flutemetamol CSV file (aibl_flutemeta_01-Jun-2018.csv)
    # has an extra column for some rows. However, each CSV file (regarding PET tracers)
    # contains the same columns. The usecols fixes this issue.
    return pd.read_csv(
        path_to_csv_pet_modality,
        sep=",|;",
        usecols=list(range(0, 36)),
        engine="python",
    )


def _check_subdirectories_pet(
    subdirectories: List[str], no_pet: List[str]
) -> List[str]:
    """Returns the correct subdirectories for the PET images, they should
    belong to the list where there all the possible names of the PET images.

    Parameters
    ----------
    subdirectories : list of str
        The possible subdirectories which need to be checked.

    no_pet : list of str
        List of folder names which do not contain PET images.

    Return
    ------
    list of str :
        List of subdirectories which contain a PET image which
        needs to be converted.
    """
    return list(
        set([s for s in subdirectories if (s not in no_pet) & (s != ".DS_Store")])
    )


def _list_folder_without_pet() -> List[str]:
    """Return a list of all the folders which do not contain PET images."""
    return [
        ".DS_Store",
        "localizer",
        "Space_3D_T2_FLAIR_sag_p2",
        "AXIAL_FLAIR",
        "MPRAGE_ADNI_confirmed_REPEATX2",
        "Axial_PD-T2_TSE",
        "Axial_PD-T2_TSE_repeat",
        "MPRAGE_SAG_ISO_p2_ND",
        "Axial_PD-T2_TSE_confirmed",
        "MPRAGESAGISOp2ND",
        "MPRAGE_ADNI_confirmed",
        "MPRAGE_ADNI_confirmed_repeat",
        "MPRAGE_SAG_ISO_p2",
        "MPRAGE",
        "MPRAGE_ADNI_confirmed_REPEAT",
        "Axial_PD-T2_TSE_confirmed_repeat",
        "MPRAGE_ADNI_conf_REPEAT",
        "Space_3D_T2_FLAIR_sag_p2_REPEAT",
        "MPRAGE_ADNI_confirmed_RPT",
        "Brain_256_1.6_zoom_4_x_4_iter",
        "Space_3D_T2_FLAIR_sag_REPEAT",
        "Axial_PD-T2_TSE_RPTconfirmed",
        "Axial_PD-T2_TSE_RPT_confirmed",
        "Axial_PD-T2_TSE_confirmed_REPEAT",
        "flair_t2_spc_irprep_ns_sag_p2_1mm_iso",
        "localiser",
    ]


def _create_file(
    image: pd.Series, modality: Modality, bids_dir: Path, overwrite: bool
) -> Optional[Path]:
    """Convert the AIBL PET images into the BIDS specification.

    There are three pet modalities: av45, pib, flute. All of them are converted in BIDS.

    Parameters
    ----------
    image : pd.Series
    modality : Modality
    bids_dir : Path
    overwrite : bool

    Return
    ------
    Path or None :
        Path to file
    """
    from clinica.cmdline import setup_clinica_logging
    from clinica.iotools.bids_utils import StudyName, bids_id_factory, json_from_dcm
    from clinica.iotools.converter_utils import viscode_to_session
    from clinica.iotools.utils.data_handling import center_nifti_origin
    from clinica.utils.stream import cprint

    # This function is executed in a multiprocessing context
    # such that we need to re-configure the clinica logger in the child processes.
    # Note that logging messages could easily be lost (for example when logging
    # to a file from two different processes). A better solution would be to
    # implement a logging process consuming logging messages from a multiprocessing.Queue...
    setup_clinica_logging("INFO")

    participant_id = bids_id_factory(StudyName.AIBL).from_original_study_id(
        image.Subjects_ID
    )
    session_id = image.Session_ID
    image_path = image[modality.name_of_path]

    try:
        image_path = Path(image_path)
    except TypeError:
        cprint(
            msg=(
                f"[{modality.upper()}] No path specified for {image.Subjects_ID} "
                f"in session {session_id}"
            ),
            lvl="info",
        )
        return None
    cprint(
        msg=(
            f"[{modality.upper()}] Processing subject {image.Subjects_ID} "
            f"in session {session_id}"
        ),
        lvl="info",
    )
    session_id = viscode_to_session(session_id)
    output_path = (
        bids_dir
        / f"{participant_id}"
        / f"{session_id}"
        / modality.bids_folder
        / f"{participant_id}_{session_id}_{modality.suffix}"
    )
    # image is saved following BIDS specifications
    if output_path.with_suffix(".nii.gz").exists() and not overwrite:
        cprint(f"Subject {image.Subjects_ID} - session {session_id} already processed.")
        output_image = output_path.with_suffix(".nii.gz")
    else:
        if output_path.with_suffix(".nii.gz").exists():
            output_path.with_suffix(".nii.gz").unlink()
        output_image = _dicom_to_nii(output_path, image_path)
        json_from_dcm(image_path, output_path.with_suffix(".json"))

    center_nifti_origin(output_image, output_image)

    return output_image


def _dicom_to_nii(output_path: Path, image_path: Path) -> Path:
    """Convert the DICOM images to NIfTI files using dcm2niix.

    Parameters
    ----------
    output_path : Path
     The path to the NIfTI image without the extension.

    image_path : Path
        The path to where the DICOM files are stored.

    Return
    ------
    Path :
        Path to the image in NIfTI format.
    """
    from clinica.iotools.bids_utils import run_dcm2niix
    from clinica.utils.stream import cprint

    try:
        output_path.parent.mkdir(parents=True)
    except OSError:
        if not output_path.parent.is_dir():
            raise
    run_dcm2niix(
        input_dir=image_path,
        output_dir=output_path.parent,
        output_fmt=output_path.name,
        compress=True,
        bids_sidecar=False,
    )
    nifti_file = output_path.with_suffix(".nii.gz")
    if not nifti_file.exists():
        cprint(f"{nifti_file} should have been created but this did not happen")

    return nifti_file



from os import PathLike
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import nibabel as nib
import numpy as np
import pandas as pd

__all__ = [
    "center_nifti_origin",
    "check_volume_location_in_world_coordinate_system",
    "check_relative_volume_location_in_world_coordinate_system",
    "center_all_nifti",
]


def center_nifti_origin(input_image: PathLike, output_image: PathLike) -> PathLike:
    """Put the origin of the coordinate system at the center of the image.

    Parameters
    ----------
    input_image : PathLike
        Path to the input image.

    output_image : PathLike
        Path to the output image (where the result will be stored).

    Returns
    -------
    PathLike :
        The path of the output image created.
    """
    input_image = nib.load(Path(input_image))
    output_image = Path(output_image)
    canonical_image = nib.as_closest_canonical(input_image)
    header = canonical_image.header
    new_image = nib.Nifti1Image(
        canonical_image.get_fdata(caching="unchanged"),
        affine=_compute_qform(header),
        header=header,
    )
    # Without deleting already-existing file, nib.save causes a severe bug on Linux system
    if output_image.is_file():
        output_image.unlink()
    nib.save(new_image, output_image)
    if not output_image.is_file():
        raise RuntimeError(
            f"NIfTI file created but Clinica could not save it to {output_image}. "
            "Please check that the output folder has the correct permissions."
        )

    return output_image


def _compute_qform(header: nib.Nifti1Header) -> np.ndarray:
    qform = np.zeros((4, 4))
    for i in range(1, 4):
        qform[i - 1, i - 1] = header["pixdim"][i]
        qform[i - 1, 3] = -1.0 * header["pixdim"][i] * header["dim"][i] / 2.0
    return qform


def check_volume_location_in_world_coordinate_system(
    nifti_list: List[PathLike],
    bids_dir: PathLike,
    modality: str = "t1w",
    skip_question: bool = False,
) -> bool:
    """Check if images are centered around the origin of the world coordinate.

    Parameters
    ----------
    nifti_list : list of PathLike
        List of path to nifti files.

    bids_dir : PathLike
        Path to bids directory associated with this check.

    modality : str, optional
        The modality of the image. Default='t1w'.

    skip_question : bool, optional
        If True, assume answer is yes. Default=False.

    Returns
    -------
    bool :
        True if they are centered, False otherwise

    Warns
    ------
    If volume is not centered on origin of the world coordinate system

    Notes
    -----
    The NIfTI file list provided in argument are approximately centered around the origin of the world coordinates.
    Otherwise, issues may arise with further processing such as SPM segmentation.
    When not centered, we warn the user of the problem propose to exit clinica to run clinica iotools center-nifti
    or to continue with the execution of the pipeline.
    """
    import click
    import numpy as np

    bids_dir = Path(bids_dir)
    list_non_centered_files = [
        Path(file) for file in nifti_list if not _is_centered(Path(file))
    ]
    if len(list_non_centered_files) == 0:
        return True
    centers = [
        _get_world_coordinate_of_center(file) for file in list_non_centered_files
    ]
    l2_norm = [np.linalg.norm(center, ord=2) for center in centers]
    click.echo(
        _build_warning_message(
            list_non_centered_files, centers, l2_norm, bids_dir, modality
        )
    )
    if not skip_question:
        _ask_for_confirmation()

    return False


def _ask_for_confirmation() -> None:
    import sys

    import click

    if not click.confirm("Do you still want to launch the pipeline?"):
        click.echo("Clinica will now exit...")
        sys.exit(0)


def _build_warning_message(
    non_centered_files: List[Path],
    centers: List[np.array],
    l2_norm: List[np.ndarray],
    bids_dir: Path,
    modality: str,
) -> str:
    warning_message = (
        f"It appears that {len(non_centered_files)} files "
        "have a center way out of the origin of the world coordinate system. SPM has a high "
        "probability to fail on these files (for co-registration or segmentation):\n\n"
    )
    df = pd.DataFrame(
        [
            (a, b, c)
            for a, b, c in zip([f.name for f in non_centered_files], centers, l2_norm)
        ],
        columns=["File", "Coordinate of center", "Distance to origin"],
    )
    warning_message += str(df)
    cmd_line = f"`clinica iotools center-nifti {bids_dir.resolve()} {bids_dir.resolve()}_centered --modality {modality}`"
    warning_message += (
        "\nIf you are trying to launch the t1-freesurfer pipeline, you can ignore this message "
        "if you do not want to run the pet-surface pipeline afterward."
    )
    warning_message += (
        "\nClinica provides a tool to counter this problem by replacing the center of the volume"
        " at the origin of the world coordinates.\nUse the following command line to correct the "
        f"header of the faulty NIFTI volumes in a new folder:\n{cmd_line}"
        "You will find more information on the command by typing "
        "clinica iotools center-nifti in the console."
    )

    return warning_message


def check_relative_volume_location_in_world_coordinate_system(
    label_1: str,
    nifti_list1: List[PathLike],
    label_2: str,
    nifti_list2: List[PathLike],
    bids_dir: PathLike,
    modality: str,
    skip_question: bool = False,
):
    """Check if the NIfTI file list `nifti_list1` and `nifti_list2` provided in argument are not too far apart,
    otherwise coreg in SPM may fail. Norm between center of volumes of 2 files must be less than 80 mm.

    Parameters
    ----------
    label_1 : str
        Label of the first nifti_list1 files (used in potential warning message).

    nifti_list1 : list of PathLike
        First list of files.

    label_2 : str
        Label of the second nifti_list.

    nifti_list2 : list of PathLike
        Second list of files, must be same length as `nifti_list1`.

    bids_dir : PathLike
        BIDS directory (used in potential warning message).

    modality : str
        String that must be used in argument of:
        clinica iotools bids --modality <MODALITY>
        (used in potential warning message).

    skip_question : bool, optional
        Disable prompts for user input (default is False)
    """
    import warnings

    from clinica.utils.stream import cprint

    warning_message = _build_warning_message_relative_volume_location(
        label_1, nifti_list1, label_2, nifti_list2, bids_dir, modality
    )
    cprint(msg=warning_message, lvl="warning")
    warnings.warn(warning_message)
    if not skip_question:
        _ask_for_confirmation()


def _build_warning_message_relative_volume_location(
    label_1: str,
    nifti_list1: List[PathLike],
    label_2: str,
    nifti_list2: List[PathLike],
    bids_dir: PathLike,
    modality: str,
) -> str:
    bids_dir = Path(bids_dir)
    file_couples = [(Path(f1), Path(f2)) for f1, f2 in zip(nifti_list1, nifti_list2)]
    df = pd.DataFrame(
        _get_problematic_pairs_with_l2_norm(file_couples),
        columns=[label_1, label_2, "Relative distance"],
    )
    if len(df) == 0:
        return
    warning_message = (
        f"It appears that {len(df)} pairs of files have an important relative offset. "
        "SPM co-registration has a high probability to fail on these files:\n\n"
    )
    warning_message += str(df)
    warning_message += (
        "\nClinica provides a tool to counter this problem by replacing the center "
        "of the volume at the origin of the world coordinates.\nUse the following "
        "command line to correct the header of the faulty NIFTI volumes in a new folder:\n\n"
        f"`clinica iotools center-nifti {bids_dir.resolve()} {bids_dir.resolve()}_centered --modality {modality}`\n\n"
        "You will find more information on the command by typing `clinica iotools center-nifti` in the console."
    )
    return warning_message


def _get_problematic_pairs_with_l2_norm(
    file_couples: List[Tuple[Path, Path]],
    threshold: float = 80.0,
) -> List[Tuple[str, str, float]]:
    l2_norm = _compute_l2_norm(file_couples)
    pairs_with_problems = [i for i, norm in enumerate(l2_norm) if norm > threshold]

    return [
        (file_couples[k][0].name, file_couples[k][1].name, l2_norm[k])
        for k in pairs_with_problems
    ]


def _compute_l2_norm(file_couples: List[Tuple[Path, Path]]) -> List[float]:
    center_coordinates = [
        (_get_world_coordinate_of_center(f[0]), _get_world_coordinate_of_center(f[1]))
        for f in file_couples
    ]

    return [np.linalg.norm(center[0] - center[1]) for center in center_coordinates]


def center_all_nifti(
    bids_dir: PathLike,
    output_dir: PathLike,
    modalities: Optional[Iterable[str]] = None,
    center_all_files: bool = False,
) -> List[Path]:
    """Center all the NIfTI images of the input BIDS folder into the empty output_dir specified in argument.

    All the files from bids_dir are copied into output_dir, then all the NIfTI images found are replaced by their
    centered version if their center is off the origin by more than 50 mm.

    Parameters
    ----------
    bids_dir : PathLike
        Path to the BIDS directory.

    output_dir : PathLike
        Path to the output directory where the centered files will be written to.

    modalities : iterable of str, optional
        Process these modalities only. Process all modalities otherwise.

    center_all_files:  bool, default=False
        Center files that may cause problem for SPM if set to False, all files otherwise.

    Returns
    -------
    list of Path
        Centered NIfTI files.
    """
    from shutil import copy, copytree

    from clinica.utils.stream import cprint

    bids_dir, output_dir = _validate_bids_and_output_dir(bids_dir, output_dir)
    for f in bids_dir.iterdir():
        if f.is_dir() and not (output_dir / f.name).is_dir():
            copytree(f, output_dir / f.name, copy_function=copy)
        elif f.is_file() and not (output_dir / f.name).is_file():
            copy(f, output_dir / f.name)
    nifti_files_filtered: List[Path] = []
    for f in output_dir.glob("**/*.nii*"):
        if modalities is None or any(
            elem.lower() in f.name.lower() for elem in modalities
        ):
            nifti_files_filtered.append(f)
    if not center_all_files:
        nifti_files_filtered = [
            file for file in nifti_files_filtered if not _is_centered(file)
        ]
    errors: List[str] = []
    for f in nifti_files_filtered:
        cprint(msg=f"Handling file {f}", lvl="debug")
        try:
            center_nifti_origin(f, f)
        except Exception as e:
            errors.append(str(e))
    if errors:
        raise RuntimeError(
            f"Clinica encountered {len(errors)} error(s) while trying to center all NIfTI images.\n"
            + "\n".join(errors)
        )
    return nifti_files_filtered


def _validate_bids_and_output_dir(
    bids_dir: PathLike, output_dir: PathLike
) -> Tuple[Path, Path]:
    from clinica.utils.exceptions import ClinicaBIDSError
    from clinica.utils.inputs import check_bids_folder

    bids_dir = Path(bids_dir)
    output_dir = Path(output_dir)
    if bids_dir == output_dir:
        raise ClinicaBIDSError(
            f"Input BIDS ({bids_dir}) and output ({output_dir}) directories must be different."
        )
    check_bids_folder(bids_dir)

    return bids_dir, output_dir


def _is_centered(nii_volume: Path, threshold_l2: int = 50) -> bool:
    """Checks if a NIfTI volume is centered on the origin of the world coordinate system.

    Parameters
    ---------
    nii_volume : Path
        The path to the NIfTI volume to check.

    threshold_l2: int, optional
        Maximum distance between the origin of the world coordinate system and the center of the volume to
        be considered centered. The threshold where SPM segmentation stops working is around 100 mm
        (it was determined empirically after several trials on a generated dataset), so default value
        is 50mm in order to have a security margin, even when dealing with co-registered files afterward.

    Returns
    -------
    bool :
        True if the volume is centered, False otherwise.

    Notes
    ------
    SPM has troubles to segment files if the center of the volume is not close from the origin of the world coordinate
    system. A series of experiment have been conducted: we take a volume whose center is on the origin of the world
    coordinate system. We add an offset using coordinates of affine matrix [0, 3], [1, 3], [2, 3] (or by modifying the
    header['srow_x'][3], header['srow_y'][3], header['srow_z'][3], this is strictly equivalent).
    It has been determined that volumes were still segmented with SPM when the L2 distance between origin and center of
    the volume did not exceed 100 mm. Above this distance, either the volume is either not segmented (SPM error), or the
    produced segmentation is wrong (not the shape of a brain anymore)
    """
    center = _get_world_coordinate_of_center(nii_volume)
    if center is None:
        raise ValueError(
            f"Unable to compute the world coordinates of center for image {nii_volume}."
            "Please verify the image data and header."
        )
    distance_from_origin = np.linalg.norm(center, ord=2)

    return distance_from_origin < threshold_l2


def _get_world_coordinate_of_center(nii_volume: Path) -> Optional[np.ndarray]:
    """Extract the world coordinates of the center of the image.

    Parameters
    ---------
    nii_volume : PathLike
        The path to the nii volume.

    Returns
    -------
    np.ndarray :
        The coordinates in the world space.

    References
    ------
    https://brainder.org/2012/09/23/the-nifti-file-format/
    """
    from nibabel.filebasedimages import ImageFileError

    from clinica.utils.exceptions import ClinicaException
    from clinica.utils.stream import cprint

    if not nii_volume.is_file():
        raise ClinicaException(
            f"The input {nii_volume} does not appear to be a path to a file."
        )
    try:
        orig_nifti = nib.load(nii_volume)
    except ImageFileError:
        cprint(
            msg=f"File {nii_volume} could not be read by nibabel. Is it a valid NIfTI file ?",
            lvl="warning",
        )
        return None

    header = orig_nifti.header
    if isinstance(header, nib.freesurfer.mghformat.MGHHeader):
        return _scale_with_affine_transformation_matrix(
            header["dims"][0:3] / 2, header, freesurfer_image=True
        )
    try:
        center_coordinates = _get_center_volume(header)
        if header["qform_code"] > 0:
            return _scale_with_rotation_matrix(center_coordinates, header)
        if header["sform_code"] > 0:
            return _scale_with_affine_transformation_matrix(center_coordinates, header)
        if header["sform_code"] == 0:
            return _scale_coordinates_by_pixdim(center_coordinates, header)
    except KeyError as e:
        raise ValueError(
            f"Cannot get the world coordinates of the center of image {nii_volume}. "
            f"This is most likely due to missing data in the header : {e}."
        )
    return None


def _get_center_volume(header: nib.Nifti1Header) -> np.ndarray:
    """Get the voxel coordinates of the center of the data, using header information.

    Parameters
    ----------
    header: Nifti1Header
        Image header which contains image metadata.

    Returns
    -------
    ndarray :
        Voxel coordinates of the center of the volume
    """
    return np.array([x / 2 for x in header["dim"][1:4]])


def _scale_with_affine_transformation_matrix(
    coordinates_vol: np.ndarray,
    header: nib.Nifti1Header,
    freesurfer_image: bool = False,
) -> np.ndarray:
    """Convert coordinates to world space using the affine transformation matrix.

    Parameters
    ----------
    coordinates_vol : ndarray
        Coordinate in the volume (raw data).

    header : Nifti1Header
        Image header containing metadata.
        The header must have sform_code > 0.

    freesurfer_image : bool, optional
        Whether the image for which the scaling is computed was obtained with
        Freesurfer (MGHImage) or not.
        Default=False.

    Returns
    -------
    ndarray
        Coordinates in the world space

    Notes
    -----
    This method is used when sform_code is larger than zero.
    It relies on a full affine matrix, stored in the header in the fields srow_[x,y,y],
    to map voxel to world coordinates. When a nifti file is created with raw data and affine=...,
    this is this method that is used to decipher the voxel-to-world correspondence.
    """
    homogeneous_coord = np.concatenate(
        (np.array(coordinates_vol), np.array([1])), axis=0
    )
    affine = (
        header.get_affine()
        if freesurfer_image
        else _get_affine_transformation_matrix(header)
    )

    return np.dot(affine, homogeneous_coord)[0:3]


def _get_affine_transformation_matrix(header: nib.Nifti1Header) -> np.ndarray:
    """Get affine transformation matrix.

    Parameters
    ----------
    header : Nifti1Header
        The image header for which to compute the affine transformation matrix.

    Returns
    -------
    ndarray :
        The computed affine transformation matrix

    References
    ----------
    https://brainder.org/2012/09/23/the-nifti-file-format/
    """
    matrix = np.zeros((4, 4))
    matrix[0, 0] = header["srow_x"][0]
    matrix[0, 1] = header["srow_x"][1]
    matrix[0, 2] = header["srow_x"][2]
    matrix[0, 3] = header["srow_x"][3]
    matrix[1, 0] = header["srow_y"][0]
    matrix[1, 1] = header["srow_y"][1]
    matrix[1, 2] = header["srow_y"][2]
    matrix[1, 3] = header["srow_y"][3]
    matrix[2, 0] = header["srow_z"][0]
    matrix[2, 1] = header["srow_z"][1]
    matrix[2, 2] = header["srow_z"][2]
    matrix[2, 3] = header["srow_z"][3]
    matrix[3, 3] = 1

    return matrix


def _scale_with_rotation_matrix(
    coordinates_vol: np.ndarray,
    header: nib.Nifti1Header,
) -> np.ndarray:
    """Convert coordinates to world space using the rotation matrix.

    Parameters
    ----------
    coordinates_vol : ndarray
        Coordinates in the volume (raw data).

    header : Nifti1Header
        Image header containing metadata.
        The header must have a qform_code > 0.

    Returns
    -------
    ndarray :
        Coordinates in the world space.

    Notes
    -----
    This method is used when short qform_code is larger than zero.
    To get the coordinates, we multiply a rotation matrix (r_mat) by coordinates_vol,
    then perform Hadamard with pixel dimension pixdim (like in method 1).
    Then we add an offset (qoffset_x, qoffset_y, qoffset_z)
    """
    from clinica.utils.stream import cprint

    q = header["pixdim"][0]
    offset = np.array([header[f"qoffset_{x}"] for x in ("x", "y", "z")])
    if q not in (-1, 1):
        cprint(
            f"Pixdim of provided header was {q} while either -1 or 1 was expected. Using 1.",
            lvl="warning",
        )
        q = 1
    return (
        np.dot(
            _get_rotation_matrix(header),
            np.array([coordinates_vol[0], coordinates_vol[1], q * coordinates_vol[2]]),
        )
        * np.array(header["pixdim"][1:4])
        + offset
    )


def _get_rotation_matrix(header: nib.Nifti1Header) -> np.ndarray:
    """Get the rotation matrix from the provided image header.

    More information here: https://brainder.org/2012/09/23/the-nifti-file-format/

    Parameters
    ----------
    header : Nifti1Header
        The header

    Returns
    -------
    np.ndarray :
        The rotation matrix.
    """
    b = header["quatern_b"]
    c = header["quatern_c"]
    d = header["quatern_d"]
    a = np.sqrt(1 - (b**2) - (c**2) - (d**2))
    rotation = np.zeros((3, 3))
    rotation[0, 0] = (a**2) + (b**2) - (c**2) - (d**2)
    rotation[0, 1] = 2 * ((b * c) - (a * d))
    rotation[0, 2] = 2 * ((b * d) + (a * c))
    rotation[1, 0] = 2 * ((b * c) + (a * d))
    rotation[1, 1] = (a**2) + (c**2) - (b**2) - (d**2)
    rotation[1, 2] = 2 * ((c * d) - (a * b))
    rotation[2, 0] = 2 * ((b * d) - (a * c))
    rotation[2, 1] = 2 * ((b * d) - (a * c))
    rotation[2, 2] = (a**2) + (d**2) - (b**2) - (c**2)

    return rotation


def _scale_coordinates_by_pixdim(
    coordinates_vol: np.ndarray,
    header: nib.Nifti1Header,
) -> np.ndarray:
    """Convert coordinates to world space by scaling with pixel dimension.

    Parameters
    ----------
    coordinates_vol: ndarray
        Coordinates in the volume (raw data).

    header: Nifti1Header
        Contains image metadata

    Returns
    -------
    ndarray
        Coordinates in the world space

    Notes
    -----
    This method is for compatibility with analyze and is not supposed to be used as the main orientation method.
    But it is used if sform_code = 0. The world coordinates are determined simply by scaling by the voxel size
    by their dimension stored in pixdim.

    References
    ----------
    https://brainder.org/2012/09/23/the-nifti-file-format/
    """
    return np.array(coordinates_vol) * np.array(
        [header["pixdim"][1], header["pixdim"][2], header["pixdim"][3]]
    )



"""This module contains Nipype tasks used by the DWIPreprocessingUsingPhaseDiff pipeline.

Nipype tasks must be 'self-contained' such that only primitive type hints can be
used. The tasks are simple wrappers around a properly implemented Python function.
"""


def rename_into_caps_task(
    dwi_filename: str,
    dwi_preproc_filename: str,
    b_values_preproc_filename: str,
    b_vectors_preproc_filename: str,
    b0_brain_mask_filename: str,
    calibrated_magnitude_image_filename: str,
    calibrated_field_map_image_filename: str,
    calibrated_smoothed_field_map_image_filename: str,
) -> tuple:
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.fmap.utils import rename_into_caps

    return rename_into_caps(
        Path(dwi_filename),
        Path(dwi_preproc_filename),
        Path(b_values_preproc_filename),
        Path(b_vectors_preproc_filename),
        Path(b0_brain_mask_filename),
        Path(calibrated_magnitude_image_filename),
        Path(calibrated_field_map_image_filename),
        Path(calibrated_smoothed_field_map_image_filename),
    )


def convert_phase_difference_to_hertz_task(
    phase_diff_filename: str,
    delta_echo_time: float,
    working_dir: str = None,
) -> str:
    """Wrapper for 'convert_phase_difference_to_hertz' to be used by Nipype."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.fmap.utils import (
        convert_phase_difference_to_hertz,
    )

    if working_dir:
        working_dir = Path(working_dir)
    return str(
        convert_phase_difference_to_hertz(
            Path(phase_diff_filename), delta_echo_time, working_dir
        )
    )


def demean_image_task(
    input_image: str, mask: str = None, working_dir: str = None
) -> str:
    """Wrapper for 'demean_image' to be used by Nipype."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.fmap.utils import demean_image

    if working_dir:
        working_dir = Path(working_dir)
    if mask:
        mask = Path(mask)
    return str(demean_image(Path(input_image), mask, working_dir))


def convert_phase_difference_to_rads_task(
    phase_diff_filename: str, working_dir: str = None
) -> str:
    """Wrapper for 'convert_phase_difference_to_rads' to be used by Nipype."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.fmap.utils import (
        convert_phase_difference_to_rads,
    )

    if working_dir:
        working_dir = Path(working_dir)
    return str(convert_phase_difference_to_rads(Path(phase_diff_filename), working_dir))



from typing import Optional

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pet-linear"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.acq_label
@cli_param.argument.suvr_reference_region
@cli_param.option_group.pipeline_specific_options
@cli_param.option.reconstruction_method
@cli_param.option.option(
    "-ui",
    "--uncropped_image",
    is_flag=True,
    help="Do not crop the image with template (cropped image are suggested for using with DL models)",
)
@cli_param.option.option(
    "--save_pet_in_t1w_space",
    is_flag=True,
    help="Save the PET image in the T1w space computed in the intermediate step of the pipeline",
)
@cli_param.option.random_seed
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    acq_label: str,
    suvr_reference_region: str,
    reconstruction_method: Optional[str] = None,
    uncropped_image: bool = False,
    save_pet_in_t1w_space: bool = False,
    random_seed: Optional[int] = None,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Affine registration of PET images to the MNI standard space.

       ACQ_LABEL corresponds the label given to the PET acquisition, specifying the tracer used.
    Frequently used values are '18FFDG' or '18FAV45'.

       The reference region must be specified to perform intensity normalization.
    Accepted values include: 'pons', 'cerebellumPons', 'pons2', 'cerebellumPons2'.

    Prerequisite: You need to have performed the t1-linear pipeline on your T1-weighted MR images.

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Linear/"
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .pipeline import PETLinear

    parameters = {
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
        "reconstruction_method": reconstruction_method,
        "uncropped_image": uncropped_image,
        "save_PETinT1w": save_pet_in_t1w_space,
        "random_seed": random_seed,
    }

    pipeline = PETLinear(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



"""This module contains utility functions which are used across Pydra workflows."""
import typing as ty


def sanitize_fwhm(
    fwhm: ty.Union[float, ty.Tuple[float], ty.List[float], ty.List[ty.List[float]]],
) -> ty.List[ty.Tuple[float, float, float]]:
    """Make sure the FWHM is in the right format for the Smooth SPM interface.

    Parameters
    ----------
    fwhm : Union[float, Tuple[float], List[float], List[List[float]]
        Smoothing kernel(s) that should get passed to the SPM Smooth() interface.
        There are three ways to specify fwhm:
            - A float
            - A list/tuple of floats
            - A list of lists of floats

    Returns
    -------
    fwhm : List[List[float]]
        The FWHM kernels as a list of lists of floats. All inner lists are of
        length 3 as they encode each physical dimension.

    Examples
    --------
    >>> _sanitize_fwhm(3.0)
    [[3.0, 3.0, 3.0]]
    >>> _sanitize_fwhm([3.0])
    [[3.0, 3.0, 3.0]]
    >>> _sanitize_fwhm((3,))
    [[3.0, 3.0, 3.0]]
    >>> _sanitize_fwhm([3.0, 2.0])
    [[3.0, 3.0, 3.0], [2.0, 2.0, 2.0]]
    >>> _sanitize_fwhm((3.0, 2.0))
    [[3.0, 3.0, 3.0], [2.0, 2.0, 2.0]]
    >>> _sanitize_fwhm([3.0, 2.0, 1.0])
    [[3.0, 3.0, 3.0], [2.0, 2.0, 2.0], [1.0, 1.0, 1.0]]
    >>> _sanitize_fwhm([[3.0, 2.0, 1.0], [2.0, 2.0, 1.0]])
    [[3.0, 2.0, 1.0], [2.0, 2.0, 1.0]]
    """
    if isinstance(fwhm, tuple):
        fwhm = list(fwhm)
    if isinstance(fwhm, (int, float)):
        fwhm = [[float(fwhm)] * 3]
    if isinstance(fwhm, list):
        if len(fwhm) == 0:
            raise ValueError("Empty FWHM list provided.")
        if isinstance(fwhm[0], list):
            if not all([isinstance(f, list) for f in fwhm]):
                raise ValueError(
                    "Expecting a list of lists of ints or a list of ints for FWHM."
                )
            if not all([len(f) == 3 for f in fwhm]):
                raise ValueError(
                    "When providing a list of lists of ints for FWHM, all inner lists must have length 3"
                )
            for f in fwhm:
                if not all([isinstance(ff, (int, float)) for ff in f]):
                    raise ValueError(
                        "Expecting a list of lists of ints or a list of ints for FWHM."
                    )
        else:
            if all([isinstance(f, (int, float)) for f in fwhm]):
                return [[float(f)] * 3 for f in fwhm]
            else:
                raise ValueError(
                    "Expecting a list of lists of floats or a list of floats for FWHM."
                )
    return fwhm



"""This module contains workflows used by the DWIPreprocessingUsingPhaseDiff pipeline."""

from typing import Optional

from nipype.pipeline.engine import Workflow

__all__ = ["compute_reference_b0", "calibrate_and_register_fmap"]


def _cleanup_edge_pipeline(name: str = "Cleanup") -> Workflow:
    """Perform some de-spiking filtering to clean up the edge of the fieldmap (copied from fsl_prepare_fieldmap).

    This workflow was taken from: https://github.com/niflows/nipype1-workflows/
    """
    import nipype.pipeline.engine as npe
    from nipype.interfaces import fsl
    from nipype.interfaces import utility as niu

    inputnode = npe.Node(
        niu.IdentityInterface(fields=["in_file", "in_mask"]), name="inputnode"
    )
    outputnode = npe.Node(niu.IdentityInterface(fields=["out_file"]), name="outputnode")
    fugue = npe.Node(
        fsl.FUGUE(
            save_fmap=True,
            despike_2dfilter=True,
            despike_threshold=2.1,
        ),
        name="Despike",
    )
    erode = npe.Node(
        fsl.maths.MathsCommand(nan2zeros=True, args="-kernel 2D -ero"),
        name="MskErode",
    )
    new_mask = npe.Node(
        fsl.MultiImageMaths(op_string="-sub %s -thr 0.5 -bin"), name="NewMask"
    )
    apply_mask = npe.Node(fsl.ApplyMask(nan2zeros=True), name="ApplyMask")
    join = npe.Node(niu.Merge(2), name="Merge")
    add_edge = npe.Node(
        fsl.MultiImageMaths(op_string="-mas %s -add %s"), name="AddEdge"
    )

    wf = npe.Workflow(name=name)

    connections = [
        (inputnode, fugue, [("in_file", "fmap_in_file"), ("in_mask", "mask_file")]),
        (inputnode, erode, [("in_mask", "in_file")]),
        (inputnode, new_mask, [("in_mask", "in_file")]),
        (erode, new_mask, [("out_file", "operand_files")]),
        (fugue, apply_mask, [("fmap_out_file", "in_file")]),
        (new_mask, apply_mask, [("out_file", "mask_file")]),
        (erode, join, [("out_file", "in1")]),
        (apply_mask, join, [("out_file", "in2")]),
        (inputnode, add_edge, [("in_file", "in_file")]),
        (join, add_edge, [("out", "operand_files")]),
        (add_edge, outputnode, [("out_file", "out_file")]),
    ]

    wf.connect(connections)

    return wf


def _prepare_phasediff_fmap(
    base_dir: str,
    output_dir: Optional[str] = None,
    name: Optional[str] = "prepare_phasediff_fmap",
) -> Workflow:
    """This workflow adapts the fsl_prepare_fieldmap script from FSL for the FSL eddy command.

    Please note that the step 3 converts the field map into Hz instead of rad/s
    in the initial script because FSL eddy --field expects the field map to be in Hz.

    This workflow needs FSL.

    The workflow performs the following steps:
        - Step 1 - Convert the field map into radians
        - Step 2 - Unwrap the field map with PRELUDE
        - Step 3 - Convert the field map to Hz
        - Step 4 - Call FUGUE to extrapolate from mask (fill holes, etc)
        - Step 5 - Demean the field map to avoid gross shifting
        - Step 6 - Clean up the edge voxels

    Parameters
    ----------
    base_dir : str
        The base directory.

    output_dir: str, optional
         Path to output directory.
         If provided, the pipeline will write its output in this folder.
         Default to None.

    name : str, optional
        Name of the workflow. Default="prepare_phasediff_fmap".

    Returns
    -------
    Workflow :
        The Nipype workflow.
         This workflow has the following inputs:
            - "fmap_mask" : str, the path to the binary mask of the field map.
            - "fmap_phasediff" : str, the path to the phase difference field map.
            - "fmap_magnitude" : str, the path to the brain extracted magnitude field map.
              Chose the field map with the best contrast.
            - "delta_echo_time" : float, the DeltaEchoTime from BIDS specifications.
        And the following outputs:
            - "calibrated_fmap" : str, the path to the calibrated field map for eddy --field command.

    Warnings
    --------
    This workflow can not be used for PRELUDE.
    You would need to change the step 3 (conversion into rad/s instead of Hz).
    """
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.io as nio
    import nipype.interfaces.utility as nutil
    import nipype.pipeline.engine as npe

    from clinica.pipelines.dwi.preprocessing.fmap.workflows import (
        _cleanup_edge_pipeline,  # noqa
    )

    from .tasks import (
        convert_phase_difference_to_hertz_task,
        convert_phase_difference_to_rads_task,
        demean_image_task,
    )

    input_node = npe.Node(
        nutil.IdentityInterface(
            fields=[
                "fmap_mask",
                "fmap_phasediff",
                "fmap_magnitude",
                "delta_echo_time",
            ]
        ),
        name="input_node",
    )

    convert_fmap_to_rads = npe.Node(
        nutil.Function(
            input_names=["phase_diff_filename"],
            output_names=["out_file"],
            function=convert_phase_difference_to_rads_task,
        ),
        name="convert_fmap_to_rads",
    )
    convert_fmap_to_rads.inputs.working_dir = base_dir

    unwrap_fmap_with_prelude = npe.Node(
        fsl.PRELUDE(process3d=True), name="unwrap_fmap_with_prelude"
    )

    convert_fmap_to_hz = npe.Node(
        nutil.Function(
            input_names=["phase_diff_filename", "delta_echo_time"],
            output_names=["out_file"],
            function=convert_phase_difference_to_hertz_task,
        ),
        name="convert_fmap_to_hz",
    )
    convert_fmap_to_hz.inputs.working_dir = base_dir

    fugue_extrapolation_from_mask = npe.Node(
        fsl.FUGUE(save_fmap=True), name="fugue_extrapolation_from_mask"
    )

    demean_fmap = npe.Node(
        nutil.Function(
            input_names=["input_image", "mask"],
            output_names=["out_file"],
            function=demean_image_task,
        ),
        name="demean_fmap",
    )
    demean_fmap.inputs.working_dir = base_dir

    cleanup_edge_voxels = _cleanup_edge_pipeline(name="cleanup_edge_voxels")

    output_node = npe.Node(
        nutil.IdentityInterface(fields=["calibrated_fmap"]),
        name="output_node",
    )

    if output_dir:
        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = output_dir
        write_results.inputs.parameterization = False

    wf = npe.Workflow(name=name)

    connections = [
        (input_node, convert_fmap_to_rads, [("fmap_phasediff", "phase_diff_filename")]),
        (convert_fmap_to_rads, unwrap_fmap_with_prelude, [("out_file", "phase_file")]),
        (input_node, unwrap_fmap_with_prelude, [("fmap_magnitude", "magnitude_file")]),
        (input_node, unwrap_fmap_with_prelude, [("fmap_mask", "mask_file")]),
        (
            unwrap_fmap_with_prelude,
            convert_fmap_to_hz,
            [("unwrapped_phase_file", "phase_diff_filename")],
        ),
        (input_node, convert_fmap_to_hz, [("delta_echo_time", "delta_echo_time")]),
        (
            convert_fmap_to_hz,
            fugue_extrapolation_from_mask,
            [("out_file", "fmap_in_file")],
        ),
        (input_node, fugue_extrapolation_from_mask, [("fmap_mask", "mask_file")]),
        (
            fugue_extrapolation_from_mask,
            demean_fmap,
            [("fmap_out_file", "input_image")],
        ),
        (input_node, demean_fmap, [("fmap_mask", "mask")]),
        (demean_fmap, cleanup_edge_voxels, [("out_file", "inputnode.in_file")]),
        (input_node, cleanup_edge_voxels, [("fmap_mask", "inputnode.in_mask")]),
        (
            cleanup_edge_voxels,
            output_node,
            [("outputnode.out_file", "calibrated_fmap")],
        ),
    ]

    if output_dir:
        connections += [
            (output_node, write_results, [("calibrated_fmap", "calibrated_fmap")]),
        ]
    wf.connect(connections)

    return wf


def compute_reference_b0(
    base_dir: str,
    b_value_threshold: float,
    use_cuda: bool,
    initrand: bool,
    output_dir: Optional[str] = None,
    name: str = "compute_reference_b0",
) -> Workflow:
    """Step 1 of the DWI preprocessing using phase diff pipeline.

    Compute the reference b0 (i.e. average b0 with EPI distortions)

    This pipeline needs:
        - MRtrix3 to compute the whole brain mask
        - FSL to run Eddy and BET

    Parameters
    ----------
    base_dir : str
        The base directory.

    b_value_threshold: float
        Threshold value to determine the B0 volumes in the DWI image

    use_cuda: bool
        Boolean to indicate whether cuda should be used or not

    initrand: bool
        ???

    output_dir: str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.

    name: str, optional
        Name of the pipeline. Default='compute_reference_b0'.

    Returns
    -------
    Workflow :
        The Nipype workflow.
        This workflow has the following inputs:
            - "dwi_filename": The path to the DWI image
            - "b_vectors_filename": The path to the associated B-vectors file
            - "b_values_filename": The path to the associated B-values file
            - "total_readout_time": The total readout time extracted from JSON metadata
            - "phase_encoding_direction": The phase encoding direction extracted from JSON metadata
            - "image_id": Prefix to be used for output files
        And the following outputs:
            - "reference_b0": The path to the compute reference B0 volume
            - "brainmask": The path to the computed whole brain mask
    """
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.io as nio
    import nipype.interfaces.mrtrix3 as mrtrix3
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as npe

    from clinica.pipelines.dwi.preprocessing.tasks import compute_average_b0_task
    from clinica.pipelines.dwi.preprocessing.workflows import eddy_fsl_pipeline

    from .utils import get_grad_fsl

    inputnode = npe.Node(
        niu.IdentityInterface(
            fields=[
                "dwi_filename",
                "b_vectors_filename",
                "b_values_filename",
                "total_readout_time",
                "phase_encoding_direction",
                "image_id",
            ]
        ),
        name="inputnode",
    )

    fsl_gradient = npe.Node(
        niu.Function(
            input_names=["b_values_filename", "b_vectors_filename"],
            output_names=["grad_fsl"],
            function=get_grad_fsl,
        ),
        name="fsl_gradient",
    )

    # Compute whole brain mask
    brain_mask = npe.Node(mrtrix3.BrainMask(), name="brain_mask")
    brain_mask.inputs.out_file = "brainmask.nii.gz"

    # Run eddy without calibrated fmap
    pre_eddy = eddy_fsl_pipeline(
        base_dir,
        use_cuda=use_cuda,
        initrand=initrand,
        image_id=True,
        compute_mask=False,
        name="pre_eddy",
    )

    # Compute the reference b0
    reference_b0 = npe.Node(
        niu.Function(
            input_names=["dwi_filename", "b_value_filename"],
            output_names=["out_b0_average"],
            function=compute_average_b0_task,
        ),
        name="reference_b0",
    )
    reference_b0.inputs.b_value_threshold = b_value_threshold

    # Compute brain mask from reference b0
    masked_reference_b0 = npe.Node(
        fsl.BET(mask=True, robust=True), name="masked_reference_b0"
    )

    outputnode = npe.Node(
        niu.IdentityInterface(fields=["reference_b0", "brainmask"]),
        name="outputnode",
    )

    if output_dir:
        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = output_dir
        write_results.inputs.parameterization = False

    wf = npe.Workflow(name=name)

    connections = [
        (
            inputnode,
            fsl_gradient,
            [
                ("b_values_filename", "b_values_filename"),
                ("b_vectors_filename", "b_vectors_filename"),
            ],
        ),
        (fsl_gradient, brain_mask, [("grad_fsl", "grad_fsl")]),
        (inputnode, brain_mask, [("dwi_filename", "in_file")]),
        (
            inputnode,
            pre_eddy,
            [
                ("total_readout_time", "inputnode.total_readout_time"),
                ("phase_encoding_direction", "inputnode.phase_encoding_direction"),
                ("dwi_filename", "inputnode.dwi_filename"),
                ("b_values_filename", "inputnode.b_values_filename"),
                ("b_vectors_filename", "inputnode.b_vectors_filename"),
                ("image_id", "inputnode.image_id"),
            ],
        ),
        (brain_mask, pre_eddy, [("out_file", "inputnode.in_mask")]),
        (inputnode, reference_b0, [("b_values_filename", "b_value_filename")]),
        (pre_eddy, reference_b0, [("outputnode.out_corrected", "dwi_filename")]),
        (reference_b0, masked_reference_b0, [("out_b0_average", "in_file")]),
        (masked_reference_b0, outputnode, [("out_file", "reference_b0")]),
        (brain_mask, outputnode, [("out_file", "brainmask")]),
    ]
    if output_dir:
        connections += [
            (outputnode, write_results, [("reference_b0", "reference_b0")]),
            (outputnode, write_results, [("brainmask", "brainmask")]),
        ]

    wf.connect(connections)

    return wf


def calibrate_and_register_fmap(
    base_dir: str,
    output_dir: Optional[str] = None,
    name: str = "calibrate_and_register_fmap",
) -> Workflow:
    """Step 2 of the DWI preprocessing using phase diff pipeline.

    Calibrate and register the field map.

    This pipeline needs:
        - ANTS
        - FSL

    Parameters
    ----------
    base_dir : str
        The base directory.

    output_dir: str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.

    name: str, optional
        Name of the pipeline. Default='calibrate_and_register_fmap'.

    Returns
    -------
    Workflow :
        The Nipype workflow.
        This workflow has the following inputs:
            - "bias_magnitude_fmap": The path to the bias magnitude field map
            - "fmap_phasediff": ??
            - "reference_b0": The path to the computed reference B0 volume.
              This is one of the outputs from the step 1 compute_reference_b0 workflow.
            - "delta_echo_time": ???
        And the following outputs:
            - "smooth_calibrated_fmap": The path to the smoothed and calibrated field map
            - "bet_magnitude_fmap_registered_onto_b0": The path to the bet magnitude field
              map registered onto the B0 volume
    """
    import nipype.interfaces.ants as ants
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.io as nio
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as npe

    from clinica.pipelines.dwi.preprocessing.fmap.workflows import (
        _prepare_phasediff_fmap,  # noqa
    )

    inputnode = npe.Node(
        niu.IdentityInterface(
            fields=[
                "bias_magnitude_fmap",
                "fmap_phasediff",
                "reference_b0",
                "delta_echo_time",
            ]
        ),
        name="inputnode",
    )

    # Bias field correction of the magnitude image
    bias_magnitude_fmap = npe.Node(
        ants.N4BiasFieldCorrection(dimension=3),
        name="bias_magnitude_fmap",
    )

    # Brain extraction of the magnitude image
    bet_magnitude_fmap = npe.Node(
        fsl.BET(frac=0.4, mask=True),
        name="bet_magnitude_fmap",
    )

    # Calibrate FMap
    calibrate_fmap = _prepare_phasediff_fmap(base_dir=base_dir, name="calibrate_fmap")

    # Register the BET magnitude fmap onto the BET b0
    register_bet_magnitude_fmap_onto_b0 = npe.Node(
        interface=fsl.FLIRT(),
        name="register_bet_magnitude_fmap_onto_b0",
    )
    register_bet_magnitude_fmap_onto_b0.inputs.dof = 6
    register_bet_magnitude_fmap_onto_b0.inputs.output_type = "NIFTI_GZ"

    # Apply the transformation on the calibrated fmap
    apply_xfm_on_calibrated_fmap = npe.Node(
        interface=fsl.ApplyXFM(),
        name="apply_xfm_on_calibrated_fmap",
    )
    apply_xfm_on_calibrated_fmap.inputs.output_type = "NIFTI_GZ"

    # Apply the transformation on the magnitude image
    apply_xfm_on_magnitude_fmap = apply_xfm_on_calibrated_fmap.clone("2e-2-MagFMapToB0")

    # Smooth the registered (calibrated) fmap
    smooth_calibrated_fmap = npe.Node(
        interface=fsl.maths.IsotropicSmooth(),
        name="smooth_calibrated_fmap",
    )
    smooth_calibrated_fmap.inputs.sigma = 4.0

    outputnode = npe.Node(
        niu.IdentityInterface(
            fields=[
                "smooth_calibrated_fmap",
                "bet_magnitude_fmap_registered_onto_b0",
                "registered_calibrated_fmap",
            ]
        ),
        name="outputnode",
    )

    if output_dir:
        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = output_dir
        write_results.inputs.parameterization = False

    wf = npe.Workflow(name=name)

    connections = [
        # Bias field correction of the magnitude image
        (inputnode, bias_magnitude_fmap, [("bias_magnitude_fmap", "input_image")]),
        # Brain extraction of the magnitude image
        (bias_magnitude_fmap, bet_magnitude_fmap, [("output_image", "in_file")]),
        # Calibration of the FMap
        (
            bet_magnitude_fmap,
            calibrate_fmap,
            [
                ("mask_file", "input_node.fmap_mask"),
                ("out_file", "input_node.fmap_magnitude"),
            ],
        ),
        (
            inputnode,
            calibrate_fmap,
            [
                ("fmap_phasediff", "input_node.fmap_phasediff"),
                ("delta_echo_time", "input_node.delta_echo_time"),
            ],
        ),
        # Register the BET magnitude fmap onto the BET b0
        (
            bet_magnitude_fmap,
            register_bet_magnitude_fmap_onto_b0,
            [("out_file", "in_file")],
        ),
        (
            inputnode,
            register_bet_magnitude_fmap_onto_b0,
            [("reference_b0", "reference")],
        ),
        # Apply the transformation on the magnitude image
        (
            register_bet_magnitude_fmap_onto_b0,
            apply_xfm_on_magnitude_fmap,
            [("out_matrix_file", "in_matrix_file")],
        ),
        (
            bias_magnitude_fmap,
            apply_xfm_on_magnitude_fmap,
            [("output_image", "in_file")],
        ),
        (inputnode, apply_xfm_on_magnitude_fmap, [("reference_b0", "reference")]),
        # Apply the transformation on the calibrated fmap
        (
            register_bet_magnitude_fmap_onto_b0,
            apply_xfm_on_calibrated_fmap,
            [("out_matrix_file", "in_matrix_file")],
        ),
        (
            calibrate_fmap,
            apply_xfm_on_calibrated_fmap,
            [("output_node.calibrated_fmap", "in_file")],
        ),
        (inputnode, apply_xfm_on_calibrated_fmap, [("reference_b0", "reference")]),
        # Smooth the registered (calibrated) fmap
        (
            apply_xfm_on_calibrated_fmap,
            smooth_calibrated_fmap,
            [("out_file", "in_file")],
        ),
        # Output connections
        (smooth_calibrated_fmap, outputnode, [("out_file", "smooth_calibrated_fmap")]),
        (
            register_bet_magnitude_fmap_onto_b0,
            outputnode,
            [("out_file", "bet_magnitude_fmap_registered_onto_b0")],
        ),
        (
            apply_xfm_on_calibrated_fmap,
            outputnode,
            [("out_file", "registered_calibrated_fmap")],
        ),
    ]

    if output_dir:
        connections += [
            (
                outputnode,
                write_results,
                [("smooth_calibrated_fmap", "smooth_calibrated_fmap")],
            ),
            (
                outputnode,
                write_results,
                [
                    (
                        "bet_magnitude_fmap_registered_onto_b0",
                        "bet_magnitude_fmap_registered_onto_b0",
                    )
                ],
            ),
            (
                outputnode,
                write_results,
                [("registered_calibrated_fmap", "registered_calibrated_fmap")],
            ),
        ]

    wf.connect(connections)

    return wf



"""This module handles Clinica exceptions."""


class ClinicaException(Exception):
    """Base class for Clinica exceptions."""


class ClinicaMissingDependencyError(ClinicaException):
    """Base class for Clinica dependencies errors."""


class ClinicaEnvironmentVariableError(ClinicaException):
    """Something is wrong with an environment variable managed by Clinica."""


class ClinicaDatasetError(ClinicaException):
    """Base class for errors related to Datasets."""


class ClinicaBIDSError(ClinicaDatasetError):
    """Base class for BIDS dataset errors."""


class ClinicaCAPSError(ClinicaDatasetError):
    """Base class for CAPS dataset errors."""


class ClinicaExistingDatasetError(ClinicaDatasetError):
    """Base class for existing dataset errors."""

    def __init__(self, dataset_folder_path):
        super().__init__(
            f"Dataset located at {dataset_folder_path} already contain some files."
        )


class ClinicaParserError(ClinicaException):
    """Base class for parser errors."""


class ClinicaXMLParserError(ClinicaParserError):
    """Base class for XML parser errors."""


class ClinicaPipelineConfigurationError(ClinicaException):
    """Base class for configuration errors of clinica pipelines."""


class ClinicaInconsistentDatasetError(ClinicaException):
    """Base class for inconsistent datasets errors."""

    def __init__(self, cross_subj: list):
        max_subjects_displayed = 50
        bound = min(len(cross_subj), max_subjects_displayed)
        self.msg = (
            f"{len(cross_subj)} subjects in your BIDS folder did not respect the longitudinal organisation "
            f"from BIDS specification.\nThe subjects concerned are (showing only the first {bound}):\n\t- "
        )
        self.msg += "\n\t- ".join(cross_subj[:bound])
        self.msg += (
            f"\nClinica does not know how to handle cross sectional dataset, but it "
            "can convert it to a Clinica compliant form (using session ses-M00)"
        )
        super(ClinicaException, self).__init__(self.msg)



from os import PathLike
from pathlib import Path
from typing import Optional

from pydra.mark import annotate, task

from clinica.utils.dwi import generate_acq_file, generate_index_file

generate_acq_file_task = task(
    annotate({"return": {"out_file": Path}})(generate_acq_file)
)
generate_index_file_task = task(
    annotate({"return": {"out_file": Path}})(generate_index_file)
)



"""This module contains dictionaries used in inputs.py::clinica_{file|group}_reader().

These dictionaries describe files to grab.
"""

import functools
from collections.abc import Iterable
from typing import Optional, Union

from clinica.pipelines.dwi.dti.utils import DTIBasedMeasure
from clinica.utils.pet import ReconstructionMethod, SUVRReferenceRegion, Tracer

# BIDS

T1W_NII = {"pattern": "sub-*_ses-*_t1w.nii*", "description": "T1w MRI"}
Flair_T2W_NII = {"pattern": "sub-*_ses-*_flair.nii*", "description": "FLAIR T2w MRI"}

# T1-FreeSurfer

T1_FS_WM = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/mri/wm.seg.mgz",
    "description": "segmentation of white matter (mri/wm.seg.mgz).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_BRAIN = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/mri/brain.mgz",
    "description": " extracted brain from T1w MRI (mri/brain.mgz).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_ORIG_NU = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/mri/orig_nu.mgz",
    "description": "intensity normalized volume generated after correction for"
    " non-uniformity in FreeSurfer (mri/orig_nu.mgz).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_LONG_ORIG_NU = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/mri/orig_nu.mgz",
    "description": "intensity normalized volume generated after correction for non-uniformity in FreeSurfer (orig_nu.mgz) in longitudinal",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer longitudinal",
}

T1_FS_WM_SURF_R = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/surf/rh.white",
    "description": "right white matter/gray matter border surface (rh.white).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_LONG_SURF_R = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/surf/rh.white",
    "description": "right white matter/gray matter border surface (rh.white) generated with t1-freesurfer-longitudinal.",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer longitudinal",
}

T1_FS_LONG_SURF_L = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/surf/lh.white",
    "description": "left white matter/gray matter border surface (lh.white) generated with t1-freesurfer-longitudinal.",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer longitudinal",
}

T1_FS_WM_SURF_L = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/surf/lh.white",
    "description": "left white matter/gray matter border surface (lh.white).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_DESTRIEUX = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/mri/aparc.a2009s+aseg.mgz",
    "description": "Destrieux-based segmentation (mri/aparc.a2009s+aseg.mgz).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_DESTRIEUX_PARC_L = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/label/lh.aparc.a2009s.annot",
    "description": "left hemisphere surface-based Destrieux parcellation (label/lh.aparc.a2009s.annot).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_LONG_DESTRIEUX_PARC_L = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/label/lh.aparc.a2009s.annot",
    "description": "left hemisphere surface-based Destrieux parcellation (label/lh.aparc.a2009s.annot) generated with t1-freesurfer-longitudinal.",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer longitudinal",
}

T1_FS_LONG_DESTRIEUX_PARC_R = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/label/rh.aparc.a2009s.annot",
    "description": "right hemisphere surface-based Destrieux parcellation (label/rh.aparc.a2009s.annot) generated with t1-freesurfer-longitudinal.",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer longitudinal",
}

T1_FS_DESTRIEUX_PARC_R = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/label/rh.aparc.a2009s.annot",
    "description": "right hemisphere surface-based Destrieux parcellation (label/rh.aparc.a2009s.annot).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_DESIKAN = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/mri/aparc+aseg.mgz",
    "description": "Desikan-based segmentation (mri/aparc.a2009s+aseg.mgz).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_DESIKAN_PARC_L = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/label/lh.aparc.annot",
    "description": "left hemisphere surface-based Desikan parcellation (label/lh.aparc.annot).",
    "needed_pipeline": "t1-freesurfer",
}

T1_FS_DESIKAN_PARC_R = {
    "pattern": "t1/freesurfer_cross_sectional/sub-*_ses-*/label/rh.aparc.annot",
    "description": "right hemisphere surface-based Desikan parcellation (label/rh.aparc.annot).",
    "needed_pipeline": "t1-freesurfer",
}

# T1-FreeSurfer-Template
T1_FS_T_DESTRIEUX = {
    "pattern": "freesurfer_unbiased_template/sub-*_long-*/mri/aparc.a2009s+aseg.mgz",
    "description": "Destrieux-based segmentation (mri/aparc.a2009s+aseg.mgz) from unbiased template.",
    "needed_pipeline": "t1-freesurfer-longitudinal or t1-freesurfer-template",
}

# T1-FreeSurfer-Longitudinal-Correction
T1_FS_LONG_DESIKAN_PARC_L = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/label/lh.aparc.annot",
    "description": "left hemisphere surface-based Desikan parcellation (label/lh.aparc.annot) generated with t1-freesurfer-longitudinal.",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer-longitudinal",
}

T1_FS_LONG_DESIKAN_PARC_R = {
    "pattern": "t1/long-*/freesurfer_longitudinal/sub-*_ses-*.long.sub-*_*/label/rh.aparc.annot",
    "description": "right hemisphere surface-based Desikan parcellation (label/rh.aparc.annot) generated with t1-freesurfer-longitudinal.",
    "needed_pipeline": "t1-freesurfer and t1-freesurfer-longitudinal",
}

T1W_LINEAR = {
    "pattern": "*space-MNI152NLin2009cSym_res-1x1x1_T1w.nii.gz",
    "description": "T1w image registered in MNI152NLin2009cSym space using t1-linear pipeline",
    "needed_pipeline": "t1-linear",
}

T2W_LINEAR = {
    "pattern": "*space-MNI152NLin2009cSym_res-1x1x1_T2w.nii.gz",
    "description": "T2w image registered in MNI152NLin2009cSym space using t2-linear pipeline",
    "needed_pipeline": "t2-linear",
}

FLAIR_T2W_LINEAR = {
    "pattern": "*space-MNI152NLin2009cSym_res-1x1x1_flair.nii.gz",
    "description": "T2w image registered in MNI152NLin2009cSym space using t2-linear pipeline",
    "needed_pipeline": "flair-linear",
}

T1W_LINEAR_CROPPED = {
    "pattern": "*space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_T1w.nii.gz",
    "description": "T1W Image registered using t1-linear and cropped "
    "(matrix size 169×208×179, 1 mm isotropic voxels)",
    "needed_pipeline": "t1-linear",
}

T2W_LINEAR_CROPPED = {
    "pattern": "*space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_T2w.nii.gz",
    "description": "T2W Image registered using t2-linear and cropped "
    "(matrix size 169×208×179, 1 mm isotropic voxels)",
    "needed_pipeline": "t2-linear",
}

FLAIR_T2W_LINEAR_CROPPED = {
    "pattern": "*space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_flair.nii.gz",
    "description": "T2W Image registered using t2-linear and cropped "
    "(matrix size 169×208×179, 1 mm isotropic voxels)",
    "needed_pipeline": "flair-linear",
}

T1W_EXTENSIVE = {
    "pattern": "*space-Ixi549Space_desc-SkullStripped_T1w.nii.gz",
    "description": "T1w image skull-stripped registered in Ixi549Space space using clinicaDL preprocessing pipeline",
    "needed_pipeline": "t1-extensive",
}

T1W_TO_MNI_TRANSFORM = {
    "pattern": "*space-MNI152NLin2009cSym_res-1x1x1_affine.mat",
    "description": "Transformation matrix from T1W image to MNI space using t1-linear pipeline",
    "needed_pipeline": "t1-linear",
}

T2W_TO_MNI_TRANSFROM = {
    "pattern": "*space-MNI152NLin2009cSym_res-1x1x1_affine.mat",
    "description": "Transformation matrix from T2W image to MNI space using t2-linear pipeline",
    "needed_pipeline": "t2-linear",
}

FLAIR_T2W_TO_MNI_TRANSFROM = {
    "pattern": "*space-MNI152NLin2009cSym_res-1x1x1_affine.mat",
    "description": "Transformation matrix from T2W image to MNI space using t2-linear pipeline",
    "needed_pipeline": "flair-linear",
}


def aggregator(func):
    """If the decorated function receives iterable arguments,
    this decorator will call the decorated function for each
    value in the iterable and aggregate the results in a list.
    This works only if the iterables provided have the same length.
    Arguments lefts as non-iterable will be repeated.

    Examples
    --------
    The function `t1_volume_native_tpm` expects an integer defining the
    mask tissue and returns a dictionary describing the files to read:

    >>> import json
    >>> from clinica.utils.input_files import t1_volume_native_tpm
    >>> print(json.dumps(t1_volume_native_tpm(1), indent=3))
    {
        "pattern": "t1/spm/segmentation/native_space/*_*_T1w_segm-graymatter_probability.nii*",
        "description": "Tissue probability map graymatter in native space",
        "needed_pipeline": "t1-volume-tissue-segmentation"
    }

    Without the `aggregator` decorator, querying files for multiple
    tissues would have to be implemented in a loop:

    >>> print(json.dumps([t1_volume_native_tpm(tissue) for tissue in (1, 2)], indent=3))
    [
        {
            "pattern": "t1/spm/segmentation/native_space/*_*_T1w_segm-graymatter_probability.nii*",
            "description": "Tissue probability map graymatter in native space",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        },
        {
            "pattern": "t1/spm/segmentation/native_space/*_*_T1w_segm-whitematter_probability.nii*",
            "description": "Tissue probability map whitematter in native space",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        }
    ]

    Although this is fine, you might not know in a pipeline what was provided (scalar or iterable).
    With the `aggregator` decorator, you can pass both:

    >>> t1_volume_native_tpm((1, 2))
    [
        {
            "pattern": "t1/spm/segmentation/native_space/*_*_T1w_segm-graymatter_probability.nii*",
            "description": "Tissue probability map graymatter in native space",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        },
        {
            "pattern": "t1/spm/segmentation/native_space/*_*_T1w_segm-whitematter_probability.nii*",
            "description": "Tissue probability map whitematter in native space",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        }
    ]

    This works also with multiple args and kwargs:

    >>> from clinica.utils.input_files import t1_volume_native_tpm_in_mni
    >>> print(json.dumps(t1_volume_native_tpm_in_mni(1, False), indent=3))
    {
        "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability.nii*",
        "description": "Tissue probability map graymatter based on native MRI in MNI space (Ixi549) without modulation.",
        "needed_pipeline": "t1-volume-tissue-segmentation"
    }
    >>> print(json.dumps(t1_volume_native_tpm_in_mni(1, (True, False)), indent=3))
    [
        {
            "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-graymatter_space-Ixi549Space_modulated-on_probability.nii*",
            "description": "Tissue probability map graymatter based on native MRI in MNI space (Ixi549) with modulation.",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        },
        {
            "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-graymatter_space-Ixi549Space_modulated-off_probability.nii*",
            "description": "Tissue probability map graymatter based on native MRI in MNI space (Ixi549) without modulation.",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        }
    ]
    >>> print(json.dumps(t1_volume_native_tpm_in_mni((1, 2), (True, False)), indent=3))
    [
        {
            "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-graymatter_space-Ixi549Space_modulated-on_probability.nii*",
            "description": "Tissue probability map graymatter based on native MRI in MNI space (Ixi549) with modulation.",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        },
        {
            "pattern": "t1/spm/segmentation/normalized_space/*_*_T1w_segm-whitematter_space-Ixi549Space_modulated-off_probability.nii*",
            "description": "Tissue probability map whitematter based on native MRI in MNI space (Ixi549) without modulation.",
            "needed_pipeline": "t1-volume-tissue-segmentation"
        }
    ]
    """

    @functools.wraps(func)
    def wrapper_aggregator(*args, **kwargs):
        # Get the lengths of iterable args and kwargs
        arg_sizes = [
            len(arg)
            for arg in args
            if (isinstance(arg, Iterable) and not isinstance(arg, str))
        ]
        arg_sizes += [
            len(arg)
            for k, arg in kwargs.items()
            if (isinstance(arg, Iterable) and not isinstance(arg, str))
        ]

        # If iterable args/kwargs have different lengths, raise
        if len(set(arg_sizes)) > 1:
            raise ValueError(f"Arguments must have the same length.")

        # No iterable case, just call the function
        if len(arg_sizes) == 0:
            return func(*args, **kwargs)

        # Handle args first by repeating non-iterable values
        arg_size = arg_sizes[0]
        new_args = []
        for arg in args:
            if not (isinstance(arg, Iterable) and not isinstance(arg, str)):
                new_args.append((arg,) * arg_size)
            else:
                new_args.append(arg)

        # Same thing for kwargs
        new_kwargs = [{} for _ in range(arg_size)]
        for k, arg in kwargs.items():
            for i in range(len(new_kwargs)):
                if not (isinstance(arg, Iterable) and not isinstance(arg, str)):
                    new_kwargs[i][k] = arg
                else:
                    new_kwargs[i][k] = arg[i]

        # Properly encapsulate in a for loop
        if len(new_args) == 0:
            return [func(**x) for x in new_kwargs]
        elif len(new_kwargs) == 0:
            return [func(*x) for x in zip(*new_args)]
        return [func(*x, **y) for x, y in zip(zip(*new_args), new_kwargs)]

    return wrapper_aggregator


@aggregator
def t1_volume_native_tpm(tissue_number: int) -> dict:
    from pathlib import Path

    from .spm import get_spm_tissue_from_index

    tissue = get_spm_tissue_from_index(tissue_number)
    return {
        "pattern": Path("t1")
        / "spm"
        / "segmentation"
        / "native_space"
        / f"*_*_T1w_segm-{tissue.value}_probability.nii*",
        "description": f"Tissue probability map {tissue.value} in native space",
        "needed_pipeline": "t1-volume-tissue-segmentation",
    }


@aggregator
def t1_volume_dartel_input_tissue(tissue_number: int) -> dict:
    from pathlib import Path

    from .spm import get_spm_tissue_from_index

    tissue = get_spm_tissue_from_index(tissue_number)
    return {
        "pattern": Path("t1")
        / "spm"
        / "segmentation"
        / "dartel_input"
        / f"*_*_T1w_segm-{tissue.value}_dartelinput.nii*",
        "description": f"Dartel input for tissue probability map {tissue.value} from T1w MRI",
        "needed_pipeline": "t1-volume-tissue-segmentation",
    }


@aggregator
def t1_volume_native_tpm_in_mni(tissue_number: int, modulation: bool) -> dict:
    from pathlib import Path

    from .spm import get_spm_tissue_from_index

    tissue = get_spm_tissue_from_index(tissue_number)
    pattern_modulation = "on" if modulation else "off"
    description_modulation = "with" if modulation else "without"

    return {
        "pattern": Path("t1")
        / "spm"
        / "segmentation"
        / "normalized_space"
        / f"*_*_T1w_segm-{tissue.value}_space-Ixi549Space_modulated-{pattern_modulation}_probability.nii*",
        "description": (
            f"Tissue probability map {tissue.value} based on "
            f"native MRI in MNI space (Ixi549) {description_modulation} modulation."
        ),
        "needed_pipeline": "t1-volume-tissue-segmentation",
    }


def t1_volume_template_tpm_in_mni(
    group_label: str, tissue_number: int, modulation: bool, fwhm: Optional[int] = None
) -> dict:
    """Build the dictionary required by clinica_file_reader to get the tissue
    probability maps based on group template in MNI space.

    Parameters
    ----------
    group_label : str
        Label used for the group of interest.

    tissue_number : int
        An integer defining the tissue of interest.

    modulation : {"on", "off"}
        Whether modulation is on or off.

    fwhm : int, optional
        The smoothing kernel in millimeters.

    Returns
    -------
    dict :
        Information dict to be passed to clinica_file_reader.
    """
    from pathlib import Path

    from .spm import get_spm_tissue_from_index

    tissue = get_spm_tissue_from_index(tissue_number)
    pattern_modulation = "on" if modulation else "off"
    description_modulation = "with" if modulation else "without"
    fwhm_key_value = f"_fwhm-{fwhm}mm" if fwhm else ""
    fwhm_description = f"with {fwhm}mm smoothing" if fwhm else "with no smoothing"

    return {
        "pattern": Path("t1")
        / "spm"
        / "dartel"
        / f"group-{group_label}"
        / f"*_T1w_segm-{tissue.value}_space-Ixi549Space_modulated-{pattern_modulation}{fwhm_key_value}_probability.nii*",
        "description": (
            f"Tissue probability map {tissue.value} based on {group_label} template in MNI space "
            f"(Ixi549) {description_modulation} modulation and {fwhm_description}."
        ),
        "needed_pipeline": "t1-volume",
    }


def t1_volume_deformation_to_template(group_label):
    from pathlib import Path

    information = {
        "pattern": Path("t1")
        / "spm"
        / "dartel"
        / f"group-{group_label}"
        / f"sub-*_ses-*_T1w_target-{group_label}_transformation-forward_deformation.nii*",
        "description": f"Deformation from native space to group template {group_label} space.",
        "needed_pipeline": "t1-volume-create-dartel",
    }
    return information


@aggregator
def t1_volume_i_th_iteration_group_template(group_label, i):
    from pathlib import Path

    information = {
        "pattern": Path(f"group-{group_label}")
        / "t1"
        / f"group-{group_label}_iteration-{i}_template.nii*",
        "description": f"Iteration #{i} of Dartel template {group_label}",
        "needed_pipeline": "t1-volume or t1-volume-create-dartel",
    }
    return information


def t1_volume_final_group_template(group_label):
    from pathlib import Path

    information = {
        "pattern": Path(f"group-{group_label}")
        / "t1"
        / f"group-{group_label}_template.nii*",
        "description": f"T1w template file of group {group_label}",
        "needed_pipeline": "t1-volume or t1-volume-create-dartel",
    }
    return information


def custom_group(pattern, description):
    information = {"pattern": pattern, "description": description}
    return information


""" DWI """

# BIDS

DWI_NII = {"pattern": "dwi/sub-*_ses-*_dwi.nii*", "description": "DWI NIfTI"}

DWI_JSON = {"pattern": "dwi/sub-*_ses-*_dwi.json", "description": "DWI JSON file"}

DWI_BVAL = {"pattern": "dwi/sub-*_ses-*_dwi.bval", "description": "bval files"}

DWI_BVEC = {"pattern": "dwi/*_dwi.bvec", "description": "bvec files"}

FMAP_PHASEDIFF_JSON = {
    "pattern": "fmap/sub-*_ses-*_phasediff.json",
    "description": "phasediff JSON file",
}

FMAP_PHASEDIFF_NII = {
    "pattern": "fmap/sub-*_ses-*_phasediff.nii*",
    "description": "phasediff NIfTI volume",
}

FMAP_MAGNITUDE1_NII = {
    "pattern": "fmap/sub-*_ses-*_magnitude1.nii*",
    "description": "magnitude1 file",
}

# CAPS

DWI_PREPROC_NII = {
    "pattern": "dwi/preprocessing/sub-*_ses-*_space-*_desc-preproc_dwi.nii*",
    "description": "preprocessed DWI",
    "needed_pipeline": "dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap",
}

DWI_PREPROC_BRAINMASK = {
    "pattern": "dwi/preprocessing/sub-*_ses-*_space-*_brainmask.nii*",
    "description": "b0 brainmask",
    "needed_pipeline": "dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap",
}

DWI_PREPROC_BVEC = {
    "pattern": "dwi/preprocessing/sub-*_ses-*_space-*_desc-preproc_dwi.bvec",
    "description": "preprocessed bvec",
    "needed_pipeline": "dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap",
}

DWI_PREPROC_BVAL = {
    "pattern": "dwi/preprocessing/*_space-*_desc-preproc_dwi.bval",
    "description": "preprocessed bval",
    "needed_pipeline": "dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap",
}


def dwi_dti(measure: Union[str, DTIBasedMeasure], space: Optional[str] = None) -> dict:
    """Return the query dict required to capture DWI DTI images.

    Parameters
    ----------
    measure : DTIBasedMeasure or str
        The DTI based measure to consider.

    space : str, optional
        The space to consider.
        By default, all spaces are considered (i.e. '*' is used in regexp).

    Returns
    -------
    dict :
        The query dictionary to get DWI DTI images.
    """
    measure = DTIBasedMeasure(measure)
    space = space or "*"

    return {
        "pattern": f"dwi/dti_based_processing/*/*_space-{space}_{measure.value}.nii.gz",
        "description": f"DTI-based {measure.value} in space {space}.",
        "needed_pipeline": "dwi_dti",
    }


""" PET """

# BIDS


def bids_pet_nii(
    tracer: Optional[Union[str, Tracer]] = None,
    reconstruction: Optional[Union[str, ReconstructionMethod]] = None,
) -> dict:
    """Return the query dict required to capture PET scans.

    Parameters
    ----------
    tracer : Tracer, optional
        If specified, the query will only match PET scans acquired
        with the requested tracer.
        If None, the query will match all PET sans independently of
        the tracer used.

    reconstruction : ReconstructionMethod, optional
        If specified, the query will only match PET scans reconstructed
        with the specified method.
        If None, the query will match all PET scans independently of the
        reconstruction method used.

    Returns
    -------
    dict :
        The query dictionary to get PET scans.
    """
    from pathlib import Path

    description = f"PET data"
    trc = ""
    if tracer is not None:
        tracer = Tracer(tracer)
        trc = f"_trc-{tracer.value}"
        description += f" with {tracer.value} tracer"
    rec = ""
    if reconstruction is not None:
        reconstruction = ReconstructionMethod(reconstruction)
        rec = f"_rec-{reconstruction.value}"
        description += f" and reconstruction method {reconstruction.value}"

    return {
        "pattern": Path("pet") / f"*{trc}{rec}_pet.nii*",
        "description": description,
    }


# PET-Volume


def pet_volume_normalized_suvr_pet(
    acq_label: Union[str, Tracer],
    group_label: str,
    suvr_reference_region: Union[str, SUVRReferenceRegion],
    use_brainmasked_image: bool,
    use_pvc_data: bool,
    fwhm: int = 0,
) -> dict:
    from pathlib import Path

    acq_label = Tracer(acq_label)
    region = SUVRReferenceRegion(suvr_reference_region)

    if use_brainmasked_image:
        mask_key_value = "_mask-brain"
        mask_description = "brain-masked"
    else:
        mask_key_value = ""
        mask_description = "full"

    if use_pvc_data:
        pvc_key_value = "_pvc-rbv"
        pvc_description = "using RBV method for PVC"
    else:
        pvc_key_value = ""
        pvc_description = "without PVC"

    if fwhm:
        fwhm_key_value = f"_fwhm-{fwhm}mm"
        fwhm_description = f"with {fwhm}mm smoothing"
    else:
        fwhm_key_value = ""
        fwhm_description = "with no smoothing"

    suvr_key_value = f"_suvr-{region.value}"

    information = {
        "pattern": Path("pet")
        / "preprocessing"
        / f"group-{group_label}"
        / f"*_trc-{acq_label.value}_pet_space-Ixi549Space{pvc_key_value}{suvr_key_value}{mask_key_value}{fwhm_key_value}_pet.nii*",
        "description": (
            f"{mask_description} SUVR map (using {region.value} region) of {acq_label.value}-PET "
            f"{pvc_description} and {fwhm_description} in Ixi549Space space based on {group_label} DARTEL template"
        ),
        "needed_pipeline": "pet-volume",
    }
    return information


def pet_linear_nii(
    acq_label: Union[str, Tracer],
    suvr_reference_region: Union[str, SUVRReferenceRegion],
    uncropped_image: bool,
) -> dict:
    from pathlib import Path

    acq_label = Tracer(acq_label)
    region = SUVRReferenceRegion(suvr_reference_region)

    if uncropped_image:
        description = ""
    else:
        description = "_desc-Crop"

    information = {
        "pattern": Path("pet_linear")
        / f"*_trc-{acq_label.value}_pet_space-MNI152NLin2009cSym{description}_res-1x1x1_suvr-{region.value}_pet.nii.gz",
        "description": "",
        "needed_pipeline": "pet-linear",
    }
    return information


# CUSTOM
def custom_pipeline(pattern, description):
    information = {"pattern": pattern, "description": description}
    return information



from pathlib import Path
from typing import List, Union

import pandas as pd

from clinica.utils.pet import SUVRReferenceRegion, Tracer

__all__ = [
    "get_suvr_mask",
    "read_psf_information",
]


def get_suvr_mask(region: Union[str, SUVRReferenceRegion]) -> Path:
    """Returns the path to the SUVR mask from SUVR reference region label.

    Parameters
    ----------
    region : str or SUVRReferenceRegion
        The label of the SUVR reference region.
        Supported labels are: 'pons', 'cerebellumPons', 'pons2', and 'cerebellumPons2'

    Returns
    -------
    Path :
        The path to the SUVR mask.
    """
    masks_dir = Path(__file__).resolve().parents[2] / "resources" / "masks"

    return masks_dir / _get_suvr_reference_region_labels_filename(
        SUVRReferenceRegion(region)
    )


def _get_suvr_reference_region_labels_filename(region: SUVRReferenceRegion) -> str:
    if region == SUVRReferenceRegion.PONS:
        return "region-pons_eroded-6mm_mask.nii.gz"
    if region == SUVRReferenceRegion.CEREBELLUM_PONS:
        return "region-cerebellumPons_eroded-6mm_mask.nii.gz"
    if region == SUVRReferenceRegion.PONS2:
        return "region-pons_remove-extrabrain_eroded-2it_mask.nii.gz"
    if region == SUVRReferenceRegion.CEREBELLUM_PONS2:
        return "region-cerebellumPons_remove-extrabrain_eroded-3it_mask.nii.gz"


def read_psf_information(
    pvc_psf_tsv_path: Path,
    subject_ids: List[str],
    session_ids: List[str],
    pet_tracer: Union[str, Tracer],
) -> List[List[int]]:
    """Read PSF information from TSV file.

    Parameters
    ----------
    pvc_psf_tsv_path : Path
        The path to the TSV file containing the following columns:
        'participant_id', 'session_id', 'acq_label', 'psf_x',
        'psf_y', and 'psf_z'

    subject_ids : List[str]
        List of participant IDs.
        ex: ['sub-CLNC01', 'sub-CLNC01']

        .. warning::
            Must have the same length as `session_ids`.

    session_ids : List[str]
        List of session IDs.
        ex: ['ses-M000', 'ses-M018']

        .. warning::
            Must have the same length as `subject_ids`.

    pet_tracer : str or Tracer
        The tracer we want to select in the 'acq_label' column.
        Other tracers will not be read in this function

    Returns
    -------
    psf : List[List[int]]
        The PSF information following [subject_ids, session_ids] order.

    Examples
    --------
    Example of pvc_psf_tsv:

    participant_id    session_id     acq_label     psf_x    psf_y    psf_z
    sub-CLNC01        ses-M000        FDG           8        9        10
    sub-CLNC01        ses-M018        FDG           8        9        10
    sub-CLNC01        ses-M000        AV45          7        6        5
    sub-CLNC02        ses-M000        FDG           8        9        10
    sub-CLNC03        ses-M000        FDG           8        9        10
    """
    pet_tracer = Tracer(pet_tracer)
    df = _read_psf_dataframe(pvc_psf_tsv_path)
    psf = []
    for subject, session in zip(subject_ids, session_ids):
        result = df.query(
            f"participant_id == '{subject}' and session_id == '{session}' and acq_label == '{pet_tracer.value}'"
        )
        if len(result) == 0:
            raise RuntimeError(
                f"Subject {subject} with session {session} and tracer {pet_tracer.value} "
                f"that you want to proceed was not found in the TSV file containing "
                f"PSF specifications ({pvc_psf_tsv_path})."
            )
        if len(result) > 1:
            raise RuntimeError(
                f"Subject {subject} with session {session} and tracer {pet_tracer.value} "
                f"that you want to proceed was found multiple times "
                f"in the TSV file containing PSF specifications ({pvc_psf_tsv_path})."
            )
        psf.append(result[["psf_x", "psf_y", "psf_z"]].values.tolist()[0])

    return psf


def _read_psf_dataframe(filename: Path) -> pd.DataFrame:
    valid_columns = {
        "participant_id",
        "session_id",
        "acq_label",
        "psf_x",
        "psf_y",
        "psf_z",
    }
    df = pd.read_csv(filename, sep="\t")
    column_mismatch = valid_columns.symmetric_difference(set(df.columns))
    if column_mismatch:
        raise IOError(
            f"The file {filename} must contain the following columns (separated by tabulations):\n"
            f"{valid_columns}\n"
            f"The provided TSV file contains the following columns instead:{list(df.columns)}\n"
            f"Pay attention to the spaces (there should be none)."
        )
    return df
